/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-14
 *
 * H&G2JCtL
 *
 * A JDBC complient BLOB. A wrapper for a streamable BLOB.
 *
 */

package org.blobstreaming.jdbc;

import org.blobstreaming.BlobStream;

import java.io.InputStream;
import java.io.OutputStream;
import java.sql.SQLException;

import com.mysql.jdbc.SQLError;
import com.mysql.jdbc.Messages;
import com.mysql.jdbc.NotImplemented;

public class Blob implements java.sql.Blob {
	com.mysql.jdbc.Connection	conn;
	byte[]						url;
	BlobStream					blob;

	Blob(com.mysql.jdbc.Connection c, byte[] url) {
		conn = c;
		this.url = url;
		blob = null;
	}

	public void free()
	{
		if (blob != null) {
			blob.close();
			blob = null;
		}
	}

	public InputStream getBinaryStream() throws SQLException {
		if (blob == null)
			blob = Util.getInternalBlob(conn, url);
		return (InputStream) blob;
	}

	public byte[] getBytes(long pos, int length) throws SQLException {
		if (pos < 1) {
			throw SQLError.createSQLException(Messages.getString("Blob.2"), //$NON-NLS-1$
					SQLError.SQL_STATE_ILLEGAL_ARGUMENT);
		}

		return Util.getBytes(conn, url, pos-1, length);
	}

	public long length() throws SQLException {
		if (blob == null)
			blob = Util.getInternalBlob(conn, url);
		return blob.getLength();
	}

	public long position(byte[] pattern, long start) throws SQLException {
		throw new NotImplemented();
	}

	public long position(java.sql.Blob pattern, long start) throws SQLException {
		return position(pattern.getBytes(0, (int) pattern.length()), start);
	}

	private void setBinaryData(byte[] newBinaryData) {
	}

	public OutputStream setBinaryStream(long indexToWriteAt) throws SQLException {
		throw new NotImplemented();
	}

	public int setBytes(long writeAt, byte[] bytes) throws SQLException {
		throw new NotImplemented();
	}

	public int setBytes(long writeAt, byte[] bytes, int offset, int length) throws SQLException {
		throw new NotImplemented();
	}

	public void truncate(long arg0) throws SQLException {
		throw new NotImplemented();
	}
}
