/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * String utilities
 *
 */

package org.blobstreaming.lib;

public class str {
	/**
	 * Index return is 0 for the first position in the string.
	 * -1 means not found.
	 */
	public static int locate(String from, String search, int count)
	{
		int i;

		if (count < 0) {
			count = -count;
			i = from.length();
			while (count > 0) {
				i = from.lastIndexOf(search, i-1);
				if (i == -1)
					break;
				count--;
			}
		}
		else {
			i = -1;
			while (count > 0) {
				i = from.indexOf(search, i+1);
				if (i == -1)
					break;
				count--;
			}
		}
		return i;
	}

	public static int locate(String from, String search)
	{
		return locate(from, search, 1);
	}

	public static String left(String from, String search, int count)
	{
		int x = locate(from, search, count);

		if (x == -1) {
			if (count >= 0)
				return from;
			return "";
		}
		return from.substring(0, x);
	}

	public static String left(String from, String search)
	{
		return left(from, search, 1);
	}

	public static String right(String from, String search, int count)
	{
		int x = locate(from, search, count);

		if (x == -1) {
			if (count >= 0)
				return "";
			return from;
		}
		return from.substring(x+search.length());
	}

	public static String right(String from, String search)
	{
		return right(from, search, 1);
	}
	
	/**
	 * Should trim CR and LF as well.
	 */
	public static String trim(String s)
	{
		return s.trim();
	}

	public static String substr(String s, int offset, int len)
	{
		return s.substring(offset, offset+len);
	}

	public static boolean listContains(String list, String val)
	{
		String item;

		while (list != "") {
			item = trim(left(list, ","));
			if (item.equals(val))
				return true;
			list = trim(right(list, ","));
		}
		return false;
	}


	public static String replaceAll(String string, String replace, String with)
	{
		StringBuffer result;

		if (string == null)
			return(null);
		result = new StringBuffer();
		while (locate(string, replace) >= 0) {
			result.append(left(string, replace));
			result.append(with);
			string = right(string, replace);
		}
		result.append(string);
		return result.toString();
	}

	public static String base64encode(String value, int line_size)
	{
		return value;
	}

	public static String base64encode(String value)
	{
		return base64encode(value, 76);
	}

	public static String base64decode(String value)
	{
		return value;
	}
}
