/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * Content Handler List
 *
 */

package org.blobstreaming.www;

import org.blobstreaming.lib.str;
import java.util.Hashtable;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;

class ContentHandlerList {
	class ContentHandlerItem {
		public String			type;
		public Class			clazz;
		public String			options;
		
		public ContentHandlerItem(String type, Class c, String options)
		{
			this.type = type;
			this.clazz = c;
			this.options = options;
		}
	}

	private Hashtable handlers;

	public ContentHandlerList()
	{
		handlers = new Hashtable();
	}

	/**
	 * A content type may containing encoding information.
	 * For example:
	 * text/html; charset=ISO-8859-4
	 * A content handler must handle various character sets
	 * and encodings.
	 * The trim() method removes the data that is not
	 * significant for identifying a content handler.
	 */
	public static String trim(String type)
	{
		return(str.trim(str.left(type, ";")).toLowerCase());
	}

	public void addHandler(String typeList, Class c, String options)
	{
		String type;

		typeList = trim(typeList);
		while (typeList != "") {
			type = str.trim(str.left(typeList, ","));
			typeList = str.right(typeList, ",");
			handlers.put(type, new ContentHandlerItem(type, c, options));
		}
	}

	protected ContentHandlerItem getItem(String type)
	{
		ContentHandlerItem	h;

		if (type == null)
			type = "application/octet-stream";
		type = trim(type);

		// 1. Handler class of type and sub-type: <type>/<sub-type>:		
		h = (ContentHandlerItem) handlers.get(type);
		if (h != null)
			return(h);

		// 2. Handler class of major type: <type>/*
		type = str.left(type, "/") + "/*";
		h = (ContentHandlerItem) handlers.get(type);
		if (h != null)
			return(h);
		
		// 3. Handler class of any type: */*
		h = (ContentHandlerItem) handlers.get("*/*");

		return(h);	
	}

	public ContentHandler getHandler(String type) throws InstantiationException, IllegalAccessException
	{
		ContentHandlerItem	i;
		Class				c;
		ContentHandler		h;
		String				options;

		i = getItem(type);
		if (i == null) {
			c = ContentHandler.class;
			options = null;
		}
		else {
			c = i.clazz;
			options = i.options;
		}
		h = (ContentHandler) c.newInstance();
		h.setType(type);
		h.setOptions(options);

		return h;
	}
}
