/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * FTP (unused)
 *
 */

package org.blobstreaming.www;

import org.blobstreaming.lib.str;

import java.io.IOException;
import java.lang.StringBuffer;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;

public class FTP extends ResourceHandler {
	protected String currentTypeCode = "A";
	
	public final String			DEFAULT_TYPE_CODE = "I";

	// Command types:
	public final int			BASIC = 0;
	public final int			UPLOAD = 1;
	public final int			DOWNLOAD = 2;

	protected Socket			dataChannel = null;

	public FTP() throws IOException
	{
		super();
		protocol = "ftp";
		setOptions("DefaultPort=21");
	}

	/**
	 * FTP Commands:
	 * USER <user-name> - Submit the users name.
	 * PASS <password> - Submit the users password.
	 * ACCT <account-information> <CRLF>
	 * CWD  <path-name> - Change directory.
	 * CDUP - Change to parent directory.
	 * SMNT <path-name> - Structure mount.
	 * QUIT - Logout and close connections.
	 * REIN - Re-initialize (logout)
	 * PORT <host-port> - Host and port for data connections.
	 *                              - h1,h2,h3,h4,p1,p2
	 * PASV - Pasive mode, the server will listen.
	 * TYPE <type-code> - (A)SCII, (I)mage [Type-Code]
	 * STRU <structure-code> - (F)ile, (R)ecord, (P)age.
	 * MODE <mode-code> - (S)tream, (B)lock, (C)ompressed
	 * RETR <path-name> - Retrieve, download. [DOWNLOAD]
	 * STOR <path-name> - Store, upload. [UPLOAD]
	 * STOU - Store unique. [UPLOAD]
	 * APPE <path-name> - Append, with create. [UPLOAD]
	 * ALLO <allocation-info> (<decimal-int>
	 *      [ R <decimal-int>]) <CRLF> - Allocate.
	 * REST <restart-marker> - Restart.
	 * RNFR <path-name> - Rename from
	 * RNTO <path-name> - Rename to
	 * ABOR - Abort command and transfer.
	 * DELE <path-name> - Delete file.
	 * RMD  <path-name>  - Remove directory.
	 * MKD  <path-name>  - Make directory.
	 * PWD   - Print Working directory.
	 * LIST [ <path-name>] - List directory. [DOWNLOAD]
	 * NLST [ <path-name>] - Name list. [DOWNLOAD]
	 * SITE <site-info> - Site specific info
	 * HELP SITE
	 * 214- The following SITE commands are recognized (* =>'s unimplemented).
	 *    UMASK   IDLE    CHMOD   HELP
	 * 214 Direct comments to ftp-bugs@suse2.snap.de.
	 * HELP UMASK
	 * 502 Unknown command UMASK.
	 * SITE HELP UMASK
	 * 214 Syntax: SITE UMASK [ <sp> umask ]
	 * SITE HELP IDLE
	 * 214 Syntax: SITE IDLE [ <sp> maximum-idle-time ]
	 * SITE HELP CHMOD
	 * 214 Syntax: SITE CHMOD <sp> mode <sp> file-name
	 * SITE HELP HELP
	 * 214 Syntax: SITE HELP [ <sp> <string> ]
	 * SYST - Server OS
	 * STAT [ <path-name>] - Server returns status
	 * HELP [ <help-info>] - Get help on <help-info>.
	 * NOOP - Do nothing (keep alive).
 	 */
	public int getCommandType(String command)
	{
		if (str.locate("STOR STOU APPE", command) >= 0)
			return(UPLOAD);
		if (str.locate("RETR LIST NLST", command) >= 0)
			return(DOWNLOAD);
		return(BASIC);
	}

	public String getUserName()
	{
		String user;
		
		user = inputProperties.getProperty("User-Name");
		if (user == null);
			user = locator.getUser();
		if (user == null)
			user = "anonymous";
		return(user);
	}

	public String getPassword()
	{
		String password;
		
		password = inputProperties.getProperty("Password");
		if (password == null);
			password = locator.getPassword();
		if (password == null)
			password = "bill.gates@microsoft.com";
		return(password);
	}

	public String getPathName()
	{
		String file;
		
		file = inputProperties.getProperty("Path-Name");
		if (file == null)
			file = locator.getFile();
		return(file);
	}

	public String getTypeCode()
	{
		String type = inputProperties.getProperty("Type-Code");

		if (type == null)
			return(DEFAULT_TYPE_CODE);
		return(type);
	}

	public String getCommandParameter(String command)
	{
		if (str.locate("CWD SMNT RETR STOR APPE DELE RMD MKD LIST STAT", command) >= 0)
			return(getPathName());
		if (str.locate("CDUP QUIT REIN PASV STOU ABOR PWD SYST NOOP", command) >= 0)
			return("");
		if (command.equals("RNFR")) {
			String file;
			
			file = inputProperties.getProperty("Rename-From");
			if (file == null)
				file = getPathName();
			return(file);
		}
		if (command.equals("RNTO"))
			return(inputProperties.getProperty("Rename-To"));
		if (command.equals("USER"))
			return(getUserName());
		if (command.equals("PASS"))
			return(getPassword());
		if (command.equals("ACCT"))
			return(inputProperties.getProperty("Account-Info"));
		if (command.equals("PORT")) {
			String host = inputProperties.getProperty("Data-Channel-Host");
			int portNo = Integer.valueOf(inputProperties.getProperty("Data-Channel-Port")).intValue();
			String port;
	
			host = str.replaceAll(host, ".", ",");
			port = (portNo / 256) + "," + (portNo % 256);

			return(host+","+port);
		}
		if (command.equals("TYPE"))
			return(getTypeCode());
		if (command.equals("STRU"))
			return(inputProperties.getProperty("Structure-Code"));
		if (command.equals("MODE"))
			return(inputProperties.getProperty("Mode-Code"));
		if (command.equals("ALLO"))
			return(inputProperties.getProperty("Allocation-Info"));
		if (command.equals("REST"))
			return(inputProperties.getProperty("Restart-Marker"));
		if (command.equals("SITE"))
			return(inputProperties.getProperty("Site-Info"));
		if (command.equals("HELP"))
			return(inputProperties.getProperty("Help-Info"));
		return("");
	}

	public String getCommandArgument(String command)
	{
		String argument;

		argument = getCommandParameter(command);
		if (argument.length() > 0) {
			return(" "+argument);
		}
		return("");
	}

	protected int readResponse(boolean haveData) throws IOException, InstantiationException, IllegalAccessException
	{
		String	result;

		result = readLine();
		if (result == null)
			result = "426 Connection closed";
		if (str.substr(result, 3, 1).equals("-")) {
			// Multi-line response.
			String code = str.substr(result, 0, 3);
			StringBuffer response = new StringBuffer();

			for (;;) {
				result = readLine();
				if (result == null)
					result = "426 Connection closed";
				if (str.locate("123456789", str.substr(result, 0, 1)) >= 0) {
					/* PMC - 1/8/2001 - 4017, some FTP servers
					 * send multiple such lines (not in FTP spec!)
					 */
					if (str.substr(result, 0, 3).equals(code) &&
						!str.substr(result, 3, 1).equals("-"))
						break;
				}
				response.append(result);
			}
			/* PMC - Do not overwrite the data when have data already! */
			if (!haveData) {
				outputContent = getContentHandler("text/plain");
				outputContent.setData(this, response.toString(), null);
			}
		}
		status = result;
		return (getStatusType(status));
	}

	private static Object portLock = new Object();
	private static int nextPort = 61234;

	public int getFreePort()
	{
		synchronized (portLock) {
			int port = nextPort;

			do {
				nextPort++;
				if (nextPort > 62234)
					nextPort = 61234;
			}
			while (!ResourceHandler.isPortFree(nextPort));
		}
		return(port);
	}

	protected int setupTransferType() throws IOException, InstantiationException, IllegalAccessException
	{
		String type = getTypeCode();

		if (currentTypeCode != type) {
			currentTypeCode = type;
			return(performCommand("TYPE"));
		}
		return(OK);
	}

	protected Socket prepareDataChannel() throws IOException, InstantiationException, IllegalAccessException
	{
		Socket	channel = null;
		int		portNo = getFreePort();

		if (setupTransferType() == OK) {
			inputProperties.setProperty("Data-Channel-Host", getLocalHost());
			inputProperties.setProperty("Data-Channel-Port", portNo);
	
			if (performCommand("PORT") == OK) {
				channel = new Socket();
	
				channel.setOptions("ListenPort="+portNo+";BlockOnListen=false");
				channel.listen();
			}
		}
		return(channel);
	}

	public int performCommand(String command) throws IOException, InstantiationException, IllegalAccessException
	{
		int		result;
		int		type;

		type = getCommandType(command);

		if (type != BASIC) {
			dataChannel = prepareDataChannel();
			if (dataChannel == null)
				return(FAILED);
		}

		//print "FTP. "+command+getCommandArgument(command);
		/* PMC - writeLine() - \r\n - is correct.
		 * writeLine(command+getCommandArgument(command));
		 * but some FTP servers do not want it!
		 *
		 * write(command+getCommandArgument(command)+str.format("\n"));
		 *
		 * PMC - I have reverted to the original implementation.
		 * The problem is (I believe), that some FTP servers
		 * want the entire line in one TCP packet!!!
		 * After version 4022, writeLine() works in this
		 * manner!
		 */
		//PBE.debug(FTP. "+command+getCommandArgument(command));
		writeLine(command+getCommandArgument(command));
		result = readResponse(false);

		if (result == EXCHANGE) {
			switch (type) {
				case UPLOAD:
					dataChannel.write(inputContent, inputContent.getLength());
					break;
				case DOWNLOAD:
					outputContent = getContentHandler("text/plain");
					outputContent.setData(this, dataChannel.getInputStream(), -1, null);
					break;
			}
			result = readResponse(type == DOWNLOAD);
		}
		else {
			if (dataChannel != null)
				dataChannel.close();
		}
	
		//PBE.debug("FTP: "+this.status);
		return(result);
	}

	public String getMethod()
	{
		String m = super.getMethod();

		if (m == null) {
			String file;

			if (inputContent != null)
				return("STOR");
			if (getPathName().endsWith("/"))
				return("LIST");
			m = "RETR";
		}
		return(m);
	}

	public int login() throws IOException, InstantiationException, IllegalAccessException
	{
		int result = OK;

		if ((result = performCommand("USER")) == MORE) {
			if ((result = performCommand("PASS")) == MORE)
				result = performCommand("ACCT");
			else if (statusCode(status) == "530")
				status = status + " (Incorrect password)";
		}
		else if (statusCode(status) == "530")
			status = status + " (User unknown)";
		return(result);
	}

	public void connect() throws IOException, InstantiationException, IllegalAccessException
	{
		super.connect();
		if (getStatusType(status) == OK) {
			status = readLine();
			if (getStatusType(status) == OK) {
				outputProperties:setProperty("Greetings", status);
				login();
			}
		}
	}

	public void syncronize() throws IOException, InstantiationException, IllegalAccessException
	{
		performCommand(getMethod());
	}

	public void close()
	{
		if (isConnected()) {
			// Ignore any error
			try {
				performCommand("QUIT");
			}
			catch (Exception e) { }
		}
		try {
			super.close();
		}
		catch (Exception e) { }
	}

	public void disconnect()
	{
		if (dataChannel != null)
			dataChannel.close();
		/* Use the super.getMethod(). We want to check if
		 * a method has been explicity specified. If this
		 * is the case, then the user is responsible for closing
		 * the connection himself.
		 */
		if (super.getMethod() == null)
			super.disconnect();
	}

	public static void register()
	{
		register(FTP.class, "ftp");
	}
}

