/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * Property List
 *
 */

package org.blobstreaming.www;

import java.util.Vector;

class PropertyList {
	class Property {
		public String	key;
		public String	value;
		
		public Property(String k, String v)
		{
			this.key = k;
			this.value = v;
		}
	}

	Vector properties;

	public PropertyList()
	{
		super();
		properties = new Vector();
	}

	/**
	 * Removes all the properties from the list.
	 */
	public void removeAll()
	{
		properties.clear();
	}

	/**
	 * Returns the number of properties in the list.
	 */
	public int length()
	{
		return(properties.size());
	}

	/**
	 * Find a property with the given key. If the key
	 * is an int then it is used as an index.
	 * null will be returned if the property is not
	 * found.
	 */
	protected Property findProperty(String key)
	{
		Property	p = null;
		int			i;

		for (i=0; i<properties.size(); i++) {
			p = (Property) properties.get(i);
			if (p.key.equalsIgnoreCase(key))
				break;
			p = null;
		}
		return p;
	}

	protected Property findProperty(int i)
	{
		Property	p = null;

		if (i < properties.size())
			p = (Property) properties.get(i);
		return p;
	}

	/**
	 * Set a property with the given key to a specified
	 * value. If the key
	 * is an int then it is used as an index.
	 */
	public void setProperty(String key, String value)
	{
		Property p;

		if ((p = findProperty(key)) != null)
			p.value = value;
		else
			properties.add(new Property(key, value));
	}

	public void setProperty(String key, int value)
	{
		setProperty(key, Integer.toString(value));
	}

	public void setProperty(String key, long value)
	{
		setProperty(key, Long.toString(value));
	}

	/**
	 * Get a property with the given key. If the key
	 * is an int then it is used as an index.
	 */
	public String getProperty(String key)
	{
		Property p;

		if ((p = findProperty(key)) != null)
			return(p.value);
		return(null);
	}

	public String getProperty(int i)
	{
		Property p;

		if ((p = findProperty(i)) != null)
			return(p.value);
		return(null);
	}

	public String getPropertyKey(int i)
	{
		Property p;

		if ((p = findProperty(i)) != null)
			return(p.key);
		return(null);
	}
}

