/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-08-30
 *
 * H&G2JCtL
 *
 * Test the Java based BLOB repository access. This test is also the
 * used to compare the performance of the file system in comparison
 * to the BLOB repository.
 */

import java.util.*;
import org.blobstreaming.www.*;
import org.blobstreaming.*;
import java.io.*;
import java.text.*;
import java.lang.Thread;

public class TestPBMS {

	//static int ITERATIONS = 1000000;
	static int ITERATIONS = 1000;

	/*
	static void upload_test()
	{
		URL	url;

		blobs = new byte[ITERATIONS][];
		try {
			url = new URL("http://localhost:8080/test/notes_tab");

			ResourceHandler resource = url.getResource();
			long start = System.currentTimeMillis();
			for (int i=0; i<ITERATIONS; i++) {
				resource.setContent("html/text", "This is an example of an uploaded BLOB", null);
				blobs[i] = resource.getContentBytes();
			}
			long stop = System.currentTimeMillis();
			printTime(start, stop);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	static void download_test()
	{
		URL	url;
		int	i = 0;

		try {
			url = new URL("http://localhost:8080/test/notes_tab");
			ResourceHandler resource = url.getResource();

			resource.setContent("html/text", "This is an example of an uploaded BLOB", null);
			String result = resource.getContentStringData();

			url = new URL("http://localhost:8080/" + result);

			long start = System.currentTimeMillis();
			for (i=0; i<ITERATIONS; i++) {
				result = url.getContentStringData();
			}
			long stop = System.currentTimeMillis();
			printTime(start, stop);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	*/


	static void printTime(long start, long stop, int iterations)
	{
		long diff = stop-start;
		DecimalFormat f = new DecimalFormat("#0.0##");
		double ops = (double) iterations / (double) diff * (double) 1000;
		StringBuffer sb = f.format(ops, new StringBuffer(), new FieldPosition(NumberFormat.INTEGER_FIELD));
		
		System.out.println("Time: "+diff+" ms, "+sb.toString()+" ops/s");
	}

	static void delete_files()
	{
		File	f;
		int		i = 0;
		boolean notFound;

		try {
			long start = System.currentTimeMillis();
			for (i=0; i<2000000; i++) {
				notFound = true;
				f = new File("tmp/blobs/blob"+i);
				if (f.exists()) {
					notFound = false;
					f.delete();
				}
				f = new File("tmp/blobs/dir"+(i%100)+"/blob"+i);
				if (f.exists()) {
					notFound = false;
					f.delete();
				}
				if (notFound)
					break;
			}
			for (int j=0; j<100; j++) {
				f = new File("tmp/blobs/dir"+j+"/");
				if (f.exists())
					f.delete();
			}
			long stop = System.currentTimeMillis();
			printTime(start, stop, i);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	static void write_file_test(boolean displayTime, boolean withSubDirs, boolean createDirs, int first, int iterations)
	{
		File				f;
		FileOutputStream	s;
		int					i = 0, num;

		try {
			if (createDirs) {
				f = new File("tmp");
				if (!f.exists())
					f.mkdir();
				f = new File("tmp/blobs");
				if (!f.exists())
					f.mkdir();
				if (withSubDirs) {
					for (i=0; i<100; i++) {
						f = new File("tmp/blobs/dir"+i+"/");
						if (!f.exists())
							f.mkdir();
					}
				}
			}


			long start = System.currentTimeMillis();
			for (i=0; i<iterations; i++) {
				num = first+i;
				if (withSubDirs)
					f = new File("tmp/blobs/dir"+(num%100)+"/blob"+num);
				else
					f = new File("tmp/blobs/blob"+num);
				f.createNewFile();
				s = new FileOutputStream(f);
				s.write("This is an example of an uploaded BLOB".getBytes());
				s.close();
			}
			long stop = System.currentTimeMillis();
			if (displayTime)
				printTime(start, stop, iterations);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	static class MultiWriteThread extends Thread {
		boolean withSubDirs;
		int start;
		int iterations;

		public MultiWriteThread(boolean w, int s, int i)
		{
			withSubDirs = w;
			start = s;
			iterations = i;
		}

		public void run() {
			write_file_test(false, withSubDirs, false, start, iterations);
		}
	}

 	static void multi_write_test(boolean withSubDirs, int threads, int total_iterations)
	{
		MultiWriteThread	t[] = new MultiWriteThread[threads];
		int					iterations;
		File				f;
		
		for (int i=0; i<100; i++) {
			f = new File("tmp/blobs/dir"+i+"/");
			if (!f.exists())
				f.mkdir();
		}

		iterations = (total_iterations+threads-1) / threads;
		for (int i=0; i<threads; i++)
			t[i] = new MultiWriteThread(withSubDirs, i * iterations, iterations);

		try {
			f = new File("tmp");
			if (!f.exists())
				f.mkdir();
			f = new File("tmp/blobs");
			if (!f.exists())
				f.mkdir();
			if (withSubDirs) {
				for (int i=0; i<100; i++) {
					f = new File("tmp/blobs/dir"+i+"/");
					if (!f.exists())
						f.mkdir();
				}
			}

			long start = System.currentTimeMillis();
			for (int i=0; i<threads; i++)
				t[i].start();
			for (int i=0; i<threads; i++)
				t[i].join();
			long stop = System.currentTimeMillis();
			printTime(start, stop, iterations * threads);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	static void read_file_test(boolean displayTime, boolean withSubDirs, int first, int iterations)
	{
		File			f;
		FileInputStream	s;
		int				i = 0, num;
		byte			buffer[] = new byte[500];
		String			str;

		try {
			long start = System.currentTimeMillis();
			for (i=0; i<iterations; i++) {
				num = first + i;
				if (withSubDirs)
					f = new File("tmp/blobs/dir"+(num%100)+"/blob"+num);
				else
					f = new File("tmp/blobs/blob"+num);
				s = new FileInputStream(f);
				s.read(buffer);
				str = new String(buffer);
				s.close();
			}
			long stop = System.currentTimeMillis();
			if (displayTime)
				printTime(start, stop, iterations);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	static class MultiReadThread extends Thread {
		boolean withSubDirs;
		int start;
		int iterations;

		public MultiReadThread(boolean w, int s, int i)
		{
			withSubDirs = w;
			start = s;
			iterations = i;
		}

		public void run() {
			read_file_test(false, withSubDirs, start, iterations);
		}
	}

 	static void multi_read_test(boolean withSubDirs, int threads, int total_iterations)
	{
		MultiReadThread	t[] = new MultiReadThread[threads];
		int				iterations;
		
		iterations = (total_iterations+threads-1) / threads;
		for (int i=0; i<threads; i++)
			t[i] = new MultiReadThread(withSubDirs, i * iterations, iterations);

		try {
			long start = System.currentTimeMillis();
			for (int i=0; i<threads; i++)
				t[i].start();
			for (int i=0; i<threads; i++)
				t[i].join();
			long stop = System.currentTimeMillis();
			printTime(start, stop, iterations * threads);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	static void upload_keep_alive_test(byte blobs[][], boolean displayTime, int first, int iterations)
	{
		try {
			Connection conn = new Connection("localhost:8080", true);

			long start = System.currentTimeMillis();
			for (int i=0; i<iterations; i++) {
				blobs[first+i] = conn.upload("/test", "html/text", "This is an example of an uploaded BLOB");
			}
			long stop = System.currentTimeMillis();
			conn.close();
			if (displayTime)
				printTime(start, stop, iterations);

		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	static class MultiUploadThread extends Thread {
		byte blobs[][];
		int start;
		int iterations;

		public MultiUploadThread(byte b[][], int s, int i)
		{
			blobs = b;
			start = s;
			iterations = i;
		}

		public void run() {
			upload_keep_alive_test(blobs, false, start, iterations);
		}
	}

 	static byte[][] multi_upload_test(int threads, int total_iterations)
	{
		MultiUploadThread	t[] = new MultiUploadThread[threads];
		byte				blobs[][];
		int					iterations;
		
		iterations = (total_iterations+threads-1) / threads;
		blobs = new byte[iterations*threads][];
		for (int i=0; i<threads; i++)
			t[i] = new MultiUploadThread(blobs, i * iterations, iterations);

		try {
			long start = System.currentTimeMillis();
			for (int i=0; i<threads; i++)
				t[i].start();
			for (int i=0; i<threads; i++)
				t[i].join();
			long stop = System.currentTimeMillis();
			printTime(start, stop, iterations * threads);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		return blobs;
	}

	static void download_keep_alive_test(byte blobs[][], boolean displayTime, int first, int iterations)
	{
		try {
			Connection conn = new Connection("localhost:8080", true);
			String result;
			BlobStream blob;

			long start = System.currentTimeMillis();
			for (int i=0; i<iterations; i++) {
				blob = conn.download(new String(blobs[first+i]));
				result = blob.getStringData();
			}
			long stop = System.currentTimeMillis();
			conn.close();
			if (displayTime)
				printTime(start, stop, iterations);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	static class MultiDownloadThread extends Thread {
		byte blobs[][];
		int start;
		int iterations;

		public MultiDownloadThread(byte b[][], int s, int i)
		{
			blobs = b;
			start = s;
			iterations = i;
		}

		public void run() {
			download_keep_alive_test(blobs, false, start, iterations);
		}
	}

 	static void multi_download_test(byte blobs[][], int threads, int total_iterations)
	{
		MultiDownloadThread t[] = new MultiDownloadThread[threads];
		int					iterations;
		
		iterations = (total_iterations+threads-1) / threads;
		for (int i=0; i<threads; i++)
			t[i] = new MultiDownloadThread(blobs, i * iterations, iterations);

		try {
			long start = System.currentTimeMillis();
			for (int i=0; i<threads; i++)
				t[i].start();
			for (int i=0; i<threads; i++)
				t[i].join();
			long stop = System.currentTimeMillis();
			printTime(start, stop, iterations * threads);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	public static void main (String args[]) {
		byte blobs[][];

		System.out.println("PBMS Repository Test");
		System.out.println("java.class.path: "+System.getProperty("java.class.path"));

		System.out.println("TEST ITERATIONS: "+ITERATIONS);

		try {
			ContentHandler.register();
			FTP.register();
			HTTP.register();

			/**
			System.out.println("REPOSITORY");
			System.out.print("Upload ");
			upload_keep_alive_test();
			System.out.print("Download ");
			download_keep_alive_test();
			//Thread.sleep(32000);
			**/
			/**
			System.out.println("FILE SYSTEM");
			System.out.print("Delete ");
			delete_files();
			System.out.print("Write ");
			write_file_test(true, false, false, 0);
			System.out.print("Read ");
			read_file_test(true, false, 0);
			**/
			/**/
			int threads = 1;
			for (int i=0; i<6; i++) {
				System.out.println("REPOSITORY, THREADS "+threads);
				System.out.print("Upload ");
				blobs = multi_upload_test(threads, ITERATIONS);
				System.out.print("Download ");
				multi_download_test(blobs, threads, ITERATIONS);
				threads *= 2;
			}
			/**/
			/**/
			threads = 1;
			for (int i=0; i<6; i++) {
				System.out.println("FILE SYSTEM, THREADS "+threads);
				System.out.print("Delete ");
				delete_files();
				System.out.print("Write ");
				multi_write_test(true, threads, ITERATIONS);
				System.out.print("Read ");
				multi_read_test(true, threads, ITERATIONS);
				threads *= 2;
			}
			/**/
		}
		catch (Exception e) {
			e.printStackTrace();
		}
    }
}
