/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 *  PrimeBase Media Stream (PBMS)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2008-09-10	Barry Leslie
 *
 * H&G2JCtL
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#ifdef DRIZZLED
#include "libdrizzle/drizzle.h"
#include "drizzle_compat.h"
#define BLOB_COL_TYPE "blob"
#else
#include "mysql/mysql.h"
#define BLOB_COL_TYPE "longblob"
#endif
#include "pbmslib.h"
//#include "libtest.h"


#define TEST_DATABASE_NAME	"test2"
#define TEST_TABLE_NAME		"PBMS_API_TEST"
#define TEST_ENGINE_NAME	"PBXT"
//#define TEST_ENGINE_NAME	"MYISAM"

#define DFLT_TEST_TIMEOUT	60


#define BUFFER_SIZE 1024
static int next_id = 1;
#define MAX_INPUT 120

//---------------------------------	
static char *get_line(pbms_bool required, char *prompt)
{
	static char input_line[MAX_INPUT];
	char *ptr, *str;
	
	do {
		printf("%s", prompt);
		
		*input_line = 0;
		ptr = fgets(input_line, MAX_INPUT, stdin);
		if (!ptr) {
			perror("fgets()");
			exit(1);
		}
		
		while (*ptr && isspace(*ptr)) ptr++;
		str = ptr;
		ptr = ptr + strlen(ptr) -1;
		
		while (*ptr && isspace(*ptr) && ptr > str) ptr--;
		ptr++;
		*ptr = 0;
	} while (required && !*str);
	
	if (!*str)
		return NULL;
		
	return str;
}

// Some metadata we can add to the BLOBs.
static char *metadata[] = {"Original-file-name", "Title", "Name", "Custom-Info", NULL};
//---------------------------------	
char *pick_metadata()
{
	char *ptr;
	int i, num =0;
	
	for (i = 0; metadata[i]; i++) {
		printf("%d: %s\n", i+1, metadata[i]);
	}

	do {
		ptr = get_line(0, "Choose a metadata field by number or hit enter for no selection:");
		if (!ptr)
			break;
		num = atoi(ptr);
		if (num > i) 
			printf("Selection out of range,\n");
		else
			break;
	} while(1);
	
	if (num)
		return metadata[--num];
	return NULL;
}
//---------------------------------	
int init_metadata(MYSQL *mysql)
{
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	int i, max, ok = 0;
	char **ptr;
	
	if (mysql_query(mysql, "SELECT * from pbms_metadata_header ;") || ((results = mysql_store_result(mysql)) == NULL)){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}
	
	max = mysql_num_rows(results);
	for (ptr = metadata; *ptr; ptr++) {
		mysql_data_seek(results, 0);
		for (i = 0; i < max; i++){
			record = mysql_fetch_row(results);
			if (!strcmp(*ptr, record[0]))
				break;
		}
		
		if (i == max) {
			char cmd[120];
			sprintf(cmd, "insert into pbms_metadata_header values(\"%s\");", *ptr);
			if (mysql_query(mysql, cmd)){
				printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
				goto done;
			}

		}
	}
	
	ok = 1;
	
done:	
	if (results)
		mysql_free_result(results);
		
	return ok;
}
	

//---------------------------------	
static int collect_metadata(PBMS pbms)
{
	char *tag, *value;
	
	pbms_clear_metadata(pbms, NULL);
	
	do {
		tag = pick_metadata();
		if (!tag)
			break;

		value = get_line(0, "Enter meta data value, if any:\n");
		if (!pbms_add_metadata(pbms, tag, (value)?value:"")) {
			printf("pbms_add_metadata() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
			return 0;
		}
	} while (1);
	
	return 1;
}

static size_t pbms_read(void *caller_data, char *buffer, size_t size)
{
	int *fh = (int*) caller_data;
	return read(*fh, buffer, size);
}

//---------------------------------	
static void insert_file(MYSQL *mysql, PBMS pbms)
{
	int fh;
	char blob_ref[MS_BLOB_URL_SIZE];
	char *file, cmd[MS_BLOB_URL_SIZE + 80];
	size_t size;
	
	file = get_line(1, "Enter the name of the file to be inserted:\n");
	fh = open(file, O_RDONLY, 0000777);
	if (fh < 0) {
		perror(file);
		return;
	}
	
	size = lseek(fh, 0, SEEK_END);
	lseek(fh, 0, SEEK_SET);
	
	// Add meta data:
	pbms_clear_metadata(pbms, NULL);
	if (!pbms_add_metadata(pbms, "Original-file-name", file)) {
		printf("pbms_add_metadata() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
		return;
	}
	
	// Send the blob data to the Media Stream engine:
//	if (!pbms_put_data_cb(pbms, TEST_TABLE_NAME, NULL, blob_ref, size, insert_file_callback, &info)) {
	if (!pbms_put_data_cb(pbms, TEST_TABLE_NAME, NULL, blob_ref, size, (PBMS_READ_CALLBACK_FUNC) pbms_read, &fh)) {
		close(fh);
		printf("pbms_put_data() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
		return;
	}
	close(fh);
	
	// Insert the blob reference into the table:
	sprintf(cmd, "insert into %s values(%d, \"%s\");", TEST_TABLE_NAME, next_id, blob_ref);
	printf("cmd: %s\n", cmd);
	
	if (mysql_query(mysql, cmd)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
	} else
		printf("File inserted with record ID: %d\n",  next_id++);
}

//---------------------------------	
static pbms_bool fetch_blob_ref(MYSQL *mysql, char *blob_ref)
{
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	char *id, cmd[80];
	
	// Step 1: Fetch the blob reference from the table:
	//--------------------------------------------------
	sprintf(cmd, "Enter the record ID of the record to be fetched. ID range = 1 to %d:\n", next_id -1);
	id = get_line(1, cmd);
	
	sprintf(cmd, "Select c2 from %s where c1 = %s;", TEST_TABLE_NAME, id);
	printf("cmd: %s\n", cmd);
	
	if (mysql_query(mysql, cmd)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		return 0;
	}
	
	results = mysql_store_result(mysql);
	if (!results){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		return 0;
	}

	if (mysql_num_rows(results) == 0) {
		mysql_free_result(results);
		printf("No records found for ID %s. \n", id);
		return 0;
	}
		
	record = mysql_fetch_row(results);
	if (strlen(record[0]) > MS_BLOB_URL_SIZE) {
		mysql_free_result(results);
		printf("A strange blob ref was returned? This should never happen! \n");
		return 0;
	}
	
	strcpy(blob_ref, record[0]);
	mysql_free_result(results);
	printf("BLOB REF: \"%s\"\n", blob_ref);
	  
	return 1;
}

//---------------------------------	
static size_t pbms_write(void *caller_data, const char *buffer, size_t size)
{
	int *fh = (int*) caller_data;	
	return write(*fh, buffer, size);
}
//---------------------------------	
static void fetch_file(MYSQL *mysql, PBMS pbms)
{
	int fh;	
	char blob_ref[MS_BLOB_URL_SIZE];
	char *file, buffer[BUFFER_SIZE];
	
	
	// Step 1: Fetch the blob reference from the table:
	//--------------------------------------------------
	if (!fetch_blob_ref(mysql, blob_ref))
		return;
	
	// Step 2: Get the blob data from the the Media Stream engine
	//         and write it to the specified file.
	//--------------------------------------------------
	file = get_line(1, "Enter the name of the file into which the BLOB is to be written:\n");
	fh = open(file, O_WRONLY | O_CREAT | O_TRUNC, 0000777);
	if (fh < 0) {
		perror(file);
		return;
	}
	
//	if (!pbms_get_data_cb(pbms, blob_ref, fetch_file_callback, &fh))
	if (!pbms_get_data_cb(pbms, blob_ref, (PBMS_WRITE_CALLBACK_FUNC) pbms_write, &fh))
		printf("pbms_get_data() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
	else
		printf("BLOB fetched OK\n");
		
	close(fh);
}

//---------------------------------	
static void insert_text(MYSQL *mysql, PBMS pbms)
{
	char blob_ref[MS_BLOB_URL_SIZE];
	char *text = NULL, *ptr, cmd[MS_BLOB_URL_SIZE + 80], *blob_ref_ptr;
	size_t space = 0, size;
	short eol =0;
	
	printf("Enter the text to be inserted, an empty line will indicate the end of the text:\n");
	while (1) {
		if (!space) {
			if (text) 
				text = (char *) realloc(text, strlen(text) + 81);
			else {
				text =  (char *) malloc(81);
				*text = 0;
			}
			space = 80;
			ptr = text + strlen(text);
			memset(ptr, 0, space +1);
		}
		*ptr = getchar();
		if (*ptr == '\n') {
			if (eol) {
				*ptr = 0;
				break;
			}
			eol = 1;
		} else
			eol = 0;
			
		ptr++;
	}
	

	size = strlen(text) +1;
	
	// Add meta data:
	if (!collect_metadata(pbms))
		return;
		
	// Send the blob data to the Media Stream engine:
	if (!pbms_put_data(pbms, TEST_TABLE_NAME, NULL, blob_ref, size, (unsigned char*)text)) {
		free(text);
		printf("pbms_put_data() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
		return;
	}
	
	blob_ref_ptr = blob_ref;
	
	// Insert the blob reference into the table:
	sprintf(cmd, "insert into %s values(%d, \"%s\");", TEST_TABLE_NAME, next_id, blob_ref);
	printf("cmd: %s\n", cmd);
	
	free(text);
	
	if (mysql_query(mysql, cmd)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
	} else
		printf("Text inserted with record ID: %d\n",  next_id++);
}


//---------------------------------	
static char *fetch_text_callback(void *caller_data, char *buffer, size_t *buffer_size, size_t size)
{
	
	// Write out the last block of data received.
	buffer[*buffer_size] = 0;
	printf("%s", buffer);
	
	if (!size) 
		return NULL; // No more date.
	
	*buffer_size = BUFFER_SIZE;
		
	return buffer;
}
//---------------------------------	
static void fetch_info(MYSQL *mysql, PBMS pbms)
{
	char blob_ref[MS_BLOB_URL_SIZE];	
	
	if (!fetch_blob_ref(mysql, blob_ref))
		return;
		
	if (pbms_get_info(pbms, blob_ref)) {
		char name[MS_META_NAME_SIZE], value[80];
		size_t value_size;
		
		pbms_reset_metadata(pbms);
		value_size = 80;
		
		printf("\nBLOB Information\n");
		printf("-----------------------------------\n");
		
		while (pbms_next_metadata(pbms, name,value, &value_size)) {
			if ((!strcmp(name, MS_CREATION_TIME)) || (!strcmp(name, MS_LAST_ACCESS))) {
				time_t secs = atol(value);
				ctime_r(&secs, value);	
			}
			
			printf("%s : %s%s\n", name, value, (value_size > 80)?"...":"");
			value_size = 80;
		}
		
		printf("-----------------------------------\n\n");
		
	} else 
		printf("pbms_get_data() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
}

//---------------------------------	
static void fetch_text(MYSQL *mysql, PBMS pbms)
{
	char blob_ref[MS_BLOB_URL_SIZE];
	char *data;
	size_t size;
	
	// Fetch the blob reference from the table:
	//--------------------------------------------------
	if (!fetch_blob_ref(mysql, blob_ref))
		return;
		
	if (!pbms_get_blob_size(pbms, blob_ref, &size)) {
		printf("pbms_get_blob_size() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
		return;
	}
	
	// The test BLOBs are relatively small so we can just allocate a buffer for the entire BLOB.
	data = (char *) malloc(size);
	pbms_get_blob_size(pbms, blob_ref, &size);
		  
	if (!pbms_get_data(pbms, blob_ref, (unsigned char*)data, size))
		printf("pbms_get_data() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
	else
		printf("BLOB data fetched:\n>>>>>>>>>>\n%s<<<<<<<<<\n", data);
		
	free(data);
		
}

//---------------------------------	
static void load_some_blobs(MYSQL *mysql, PBMS pbms, int start, int count, int min_blob, int max_blob, char *text)
{
	char cmd[MS_BLOB_URL_SIZE + 80], blob_ref[MS_BLOB_URL_SIZE];
	int i = start +1, max = max_blob - min_blob;
	size_t size;
	
	while   (count--) {				
		size = random() % max + min_blob;
		
		// Send the blob data to the Media Stream engine:
		if (!pbms_put_data(pbms, TEST_TABLE_NAME, NULL, blob_ref, size, (unsigned char*)text)) {
			printf("pbms_put_data() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
			return;
		}
				
		// Insert the blob reference into the table:
		sprintf(cmd, "insert into %s values(%d, \"%s\");", TEST_TABLE_NAME, next_id, blob_ref);
		printf("cmd: %s\n", cmd);
		
		if (mysql_query(mysql, cmd)) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			return;
		}		
	}
	
}
//---------------------------------	
static void load_blobs(MYSQL *mysql, PBMS pbms)
{
	char *text = NULL, *ptr;
	int i, blob_count, min_blob, max_blob;
	
	ptr = get_line(1, "Enter the number of BLOB to load:\n");
	blob_count = atol(ptr);
	
	ptr = get_line(1, "Enter the min BLOB size:\n");
	min_blob = atol(ptr);
	
	ptr = get_line(1, "Enter the max BLOB size:\n");
	max_blob = atol(ptr);
	
	if ((!blob_count) || (!max_blob) || (max_blob < min_blob)) {
		printf("Don't be silly!\n");
		return;
	}
	
	text = malloc(max_blob);
	for ( i = 0; i < max_blob; i++)
		text[i] = random() % 256;
	
	load_some_blobs(mysql, pbms, 0, blob_count, min_blob, max_blob, text);
	free(text);	
}



///////////////////////////////////////////////////////////////////////////
// Example of the Media Stream engine API usage;
///////////////////////////////////////////////////////////////////////////
enum test_commands {InsertFile, InsertText, FetchFile, FetchText, FetchInfo, LoadBlobs, Quit};
static enum test_commands  get_cmd();

#define TEST_BLOB_BUFFER_SIZE	100
int init_test_api(MYSQL *mysql)
{ 
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	short new_database = 0, new_table = 0, ok = 0;
	pbms_bool engine_is_pbms_enabled = 1;
	
	if (!pbms_library_init()) {
		printf("pbms_library_init() ERROR(%d): %s\n", pbms_errno(NULL), pbms_error(NULL));
		goto done;
	}

/*
	if (!pbms_enabled(mysql, TEST_ENGINE_NAME, &engine_is_pbms_enabled)) {
		printf("pbms_enabled() ERROR(%d): %s\n", pbms_errno(NULL), pbms_error(NULL));
		goto done;
	}
*/	
	if (!engine_is_pbms_enabled) {
		printf("No PBMS support for \"%s\" engine.\n", TEST_ENGINE_NAME);
		goto done;
	}
	
	if (mysql_query(mysql, "show databases like \"" TEST_DATABASE_NAME "\";")) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}
	
	results = mysql_store_result(mysql);
	if (!results){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}

	if (mysql_num_rows(results) != 1) {
		if (mysql_query(mysql, "create database " TEST_DATABASE_NAME ";")){
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			goto done;
		}
		
		new_database = 1;
	}
		
	mysql_free_result(results); results = NULL;

	if (mysql_query(mysql, "use " TEST_DATABASE_NAME ";")){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}

	// Create the table used for the test if it doesn't already exist.
	if (mysql_query(mysql, "show table status like \""  TEST_TABLE_NAME "\";")) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}

	results = mysql_store_result(mysql);
	if (!results){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}

	if (mysql_num_rows(results) == 1) {
		new_table = 0;
		do {
			char *ans;
			
			printf("The table \"%s\" already exists in the 'test' database.\n", TEST_TABLE_NAME);
			printf("I will need to delete it to run this test.\n");
			printf("Can I drop the table \"%s\" from the 'test' database? ", TEST_TABLE_NAME);
			ans = get_line(1, "[y/n]:");
			if (*ans == 'n' || *ans == 'N') {
				mysql_free_result(results);
				results = NULL;
				if (mysql_query(mysql, "SELECT max(c1) from " TEST_TABLE_NAME ";") || ((results = mysql_store_result(mysql)) == NULL)){
					printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
					goto done;
				}
				
				if (mysql_num_rows(results) == 1) {
					record = mysql_fetch_row(results);
					if (record[0])
						next_id = atol(record[0]) + 1;
				}
				
				break;
			}
	
			if (*ans == 'y' || *ans == 'Y') {
				if (mysql_query(mysql, "drop table " TEST_TABLE_NAME ";")){
					printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
					goto done;
				}
				if (!engine_is_pbms_enabled) {
			/* No longer supported
					if (!pbms_drop_table_blobs(mysql, TEST_DATABASE_NAME, TEST_TABLE_NAME)) {
						printf("pbms_drop_table_blobs() ERROR(%d): %s\n", pbms_errno(NULL), pbms_error(NULL));
						goto done;
					}
			*/
				}
				new_table = 1;
				break;
			}
		} while(1);
	} else
		new_table = 1;
		
	mysql_free_result(results);
	results = NULL;

	if (new_table) {
		if (mysql_query(mysql, "create table " TEST_TABLE_NAME "(c1 int, c2 "BLOB_COL_TYPE") engine = " TEST_ENGINE_NAME ";")){
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			goto done;
		}
		
	}
	
	
	if (!engine_is_pbms_enabled) { // At this time PBXT is the only PBMS enabled engine.
/* No longer supported
		if (!pbms_init_table(mysql, TEST_DATABASE_NAME, TEST_TABLE_NAME, 0)) {
			printf("pbms_init_table() ERROR(%d): %s\n", pbms_errno(NULL), pbms_error(NULL));
			goto done;
		}
*/
	}
	
	// Add the metadata tags we are interested in storing:
	if (!init_metadata(mysql))
		goto done;
		
	ok = 1;
	// Clean up
	done:	
	
	if (!ok) {
	
		if (new_database) {
			mysql_query(mysql, "drop database " TEST_DATABASE_NAME ";");
		} else if (new_table) {
/* No longer supported
			if ((!engine_is_pbms_enabled) && !pbms_drop_table_blobs(mysql, TEST_DATABASE_NAME, TEST_TABLE_NAME)) {
				printf("pbms_drop_table_blobs() ERROR(%d): %s\n", pbms_errno(NULL), pbms_error(NULL));
			}
*/
			mysql_query(mysql, "drop table " TEST_TABLE_NAME ";");
		}
	}
	
	if (results)
		mysql_free_result(results);
		
	return ok;

}

/*
 * my_pbms_mysql_connect() can be used to make a PBMS connection via a MySQL connection.
 */
static PBMS my_pbms_mysql_connect(MYSQL *mysql, const char *database)
{
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	PBMS pbms = NULL;
	int port;
	char *host = mysql->host;

	if (!host) {
		printf("No host name associated with the MYSQL connection handle.\n");
		goto done;
	}
	
	if (mysql_query(mysql, "show variables like  \"pbms_port\";")){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}

	results = mysql_store_result(mysql);
	if (!results){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}

	if (mysql_num_rows(results) != 1) {
		printf("No PBMS blob streaming server associated with the MYSQL connection handle.\n" );
		goto done;
	}
	
	record = mysql_fetch_row(results);
	if ((!record) || !record[1]) {
		printf("show variables like  \"pbms_port\"; Returned unexpected NULLs.\n" );
		goto done;
	}
	port = atol(record[1]);
	
	pbms = pbms_connect(host, port, database);
	if (!pbms) {
		printf( "PBMS connect failed: PBMS ERROR: %d \"%s\" line %d\n", pbms_errno(NULL), pbms_error(NULL), __LINE__);		
	}
	
done:	
	if (results)
		mysql_free_result(results);
		
	return(pbms);
}

/////////////////////////////////////////////////
void test_api(MYSQL *mysql)
{
	unsigned int port = 0;
	char *test_blob = "A test string.", *host = NULL, host_buffer[1024];
	char test_blob_buffer[TEST_BLOB_BUFFER_SIZE];
	char blob_ref[MS_BLOB_URL_SIZE];
	pbms_bool ok = 0;
	PBMS pbms = NULL;
	size_t size;
	
	if (!init_test_api(mysql))
		return;
		
	do {
			
		pbms = my_pbms_mysql_connect(mysql, TEST_DATABASE_NAME);
		if (!pbms)
			break;
		printf("Connection created via MySQL handle.\n");
		
		
		// Get some connection information.
		//--------------------------------------
		if (!pbms_get_option(pbms, PBMS_OPTION_PORT, &port)) 
			printf("pbms_get_option(PBMS_OPTION_PORT) ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
		else
			printf("PBMS_OPTION_PORT: %d\n", port);
			
		if (!pbms_get_option(pbms, PBMS_OPTION_HOST, &host)) 
			printf("pbms_get_option(PBMS_OPTION_HOST) ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
		else {
			printf("PBMS_OPTION_HOST: %s\n", host);
			strcpy(host_buffer, host); // I am assuming this will fit in host_buffer.
			host = host_buffer; // I need to do thios because host will no longer be valid after calling pbms_close().
		}
			
		// Close the connection and try connecting directly.
		// This is just for testing, there is no reason to do this otherwise.		
		//-------------------------------------------
		pbms_close(pbms);

		pbms = pbms_connect(host, port, TEST_DATABASE_NAME);
		if (!pbms)
			break;
		printf("Connection created.\n");
		
		// Send a simple blob to the Media Stream engine.
		// The engine will return a blob reference string that can be
		// use to get the blob data from the server.
		//
		// You do not explicetly delete blobs from the Media Stream engine. When first
		// uploaded the blob is created on the Media Stream engine as a temporary blob
		// and is scheduled for deletion if the blob reference has not been commited to
		// a database. The time before deletion is controlled by the  Media Stream engine's
		// pbms_temp_blob_timeout system variable.
		//---------------------------------------------------
		size =  strlen(test_blob) +1;
		if (!pbms_put_data(pbms, TEST_TABLE_NAME, NULL, blob_ref, size, (unsigned char*)test_blob))
			break;
			
		printf("PUT BLOB OK, BLOB REF: \"%s\"\n", blob_ref);
		
		if (!pbms_get_blob_size(pbms, blob_ref, &size))
			break;

		if (size != (strlen(test_blob) +1))
			printf("Unexpected blob size: %d expected %d.\n",(int) size, (int)(strlen(test_blob) +1) );
		else
			printf("Blob size looks OK.\n");
			
		// Get the blob back again.
		// The blob will remain on the Media Stream engine 
		//---------------------------------------------------

		if (!pbms_get_data(pbms, blob_ref, (unsigned char*)test_blob_buffer, TEST_BLOB_BUFFER_SIZE))
			break;
			
		if (strcmp(test_blob_buffer, test_blob))
			printf("Unexpected blob returned: \"%s\" expected: \"%s\"\n", test_blob_buffer, test_blob);
		else
			printf("Blob looks OK.\n");
			
		
		// Enter an interactive mode where the user can insert and fetch blob data.
		int running = 1;


		 while (running) {
			switch (get_cmd()) {
				case InsertFile: 
					insert_file(mysql, pbms);
					break;

				case InsertText: 
					insert_text(mysql, pbms);
					break;

				case FetchFile: 
					fetch_file(mysql, pbms);
					break;

				case FetchText: 
					fetch_text(mysql, pbms);
					break;
				
				case FetchInfo: 
					fetch_info(mysql, pbms);
					break;
					
				case LoadBlobs:
					load_blobs(mysql, pbms);
					break;
				
				
				case Quit:
					running = 0;
			}
			
		}
		
		ok = 1;
	} while(0);
	
	if (!ok) {
		printf("ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
	}else
		printf("Test completed OK.\n");
		
	if (pbms)
		pbms_close(pbms);
		
	pbms_library_end();
	
}

///////////////////////////////////////////////////////////////////////////
// Setup the test
///////////////////////////////////////////////////////////////////////////
static enum test_commands  get_cmd()
{
	char *cmd;
	printf("Commands: 'it' = Insert text as blob, 'if' = Insert file as blob, 'ft' = Fetch blob as text, 'ff' = Fetch blob as file, 'fi' = Fetch blob info, 'q' = quit\n");
	
	do {
		cmd = get_line(1, "Enter a command:\n");
		switch ( *cmd ) {
			case 'i':				
			case 'I':
				cmd++;
				if (*cmd == 't' || *cmd == 'T')
					return InsertText;
					
				if (*cmd == 'f' || *cmd == 'F')
					return InsertFile;
			break;
			
			case 'f':				
			case 'F':
				cmd++;
				if (*cmd == 't' || *cmd == 'T')
					return FetchText;
					
				if (*cmd == 'f' || *cmd == 'F')
					return FetchFile;
					
				if (*cmd == 'i' || *cmd == 'I')
					return FetchInfo;
					
			break;
			
			case 'l':				
			case 'L':
				return LoadBlobs;
				
			case 'q':
			case 'Q':
				return Quit;
		}
		
	} while(1);
	
	return Quit;
}


///////////////////////////////////////////////////////////////////////////
// Setup the test
///////////////////////////////////////////////////////////////////////////
//---------------------------------	
static void display_help(const char *app)
{
	printf("\nUsage:\n");
	printf("%s [-help] | [ -test <timeout>] [<host> [<port> [<user> [<password>]]]]\n\n", app);
	
	printf("The values for missing parameters will be taken from the default settings in\n");
	printf("the 'pbmstest.cnf.cnf' file.\n\n");
	
	printf("This is a test application for the Media Stream engine client API.\n");
	printf("The source code for this app \"%s\" is intended to provide example code.\n", __FILE__);
	printf("for developers wishing to use the API. Simply running this application\n");
	printf("itself, other than demonstrating that it works, doesn't do anything interesting.\n\n");
	
	printf("Use of the '-test' option will cause the application to run a multi threaded stress\n");
	printf("test for the number of seconds given in the '<timeout>' parameter. More than one instance of \n");
	printf("the application can be run in this mode to increase the load on teh server. The default value\n");
	printf("for '<timeout>' is %d seconds.\n\n", DFLT_TEST_TIMEOUT);
}

//---------------------------------	

int main (int argc, const char * argv[]) {
	char *host = NULL, *user = NULL, *password = NULL;
	int i, port = 0, timeout = 0, skip_test = 0;
	MYSQL *mysql = NULL;
	MYSQL_RES *results = NULL;            
	short test_mode = 0;
	
	// do some quick and dirty parameter parsing:
	if (argc > 1) {
		// Process commandline options:
		if ((argc == 2) && strcmp(argv[1], "-test")) {
			// just assume it is the 'help' option.
			display_help(argv[0]);
			return 0;
		}
		
		if (!strcmp(argv[1], "-test")) {
			test_mode = 1;
			if (argc > 2)
				timeout = atol(argv[2]);
				
			if (!timeout) {
				printf("Invalid timeout value specified for '-test' mode.\n");
				display_help(argv[0]);
				return 0;
			}
			skip_test = 2;
		}
		
		for (i = 1; i < argc  - skip_test; i++) {
			char *arg = (char *) argv[i+ skip_test];
			
			switch (i) {
				case 1:
					host = arg;
					break;
				case 2:
					port = atol(arg);
					break;
				case 3:
					user = arg;
					break;
				case 4:
					password = arg;
					break;
				default:
					display_help(argv[0]);
					return 1;

			} 
		}

	}
	//---------------------------------
	mysql = mysql_init(NULL);
	if (!mysql) {
		printf( "mysql_init() failed.\n");
		return 1;
	}
	
	if (mysql_options(mysql, MYSQL_READ_DEFAULT_FILE, "./pbmstest.cnf")){
		printf( "mysql_options() ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}

		
	printf("mysql_real_connect(mysql, \"%s\", \"%s\", \"%s\", NULL, %d , NULL, 0)\n", (host)?host: "NULL", (user)?user: "NULL", (password)?password: "NULL", port);
	if (mysql_real_connect(mysql, host, user, password, NULL, port, NULL, 0) == NULL){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}


	// Check that the Media Stream engine and the tageted Media Stream enabled engines are supported.
	if (mysql_query(mysql, "select PLUGIN_STATUS from information_schema.plugins where PLUGIN_NAME = \"PBMS\";") || ((results = mysql_store_result(mysql)) == NULL)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}
	
	mysql_free_result(results); results = NULL;

	if (mysql_query(mysql, "select PLUGIN_STATUS from information_schema.plugins where PLUGIN_NAME = \"" TEST_ENGINE_NAME "\";") || ((results = mysql_store_result(mysql)) == NULL)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		goto done;
	}
	
	
	mysql_free_result(results); results = NULL;

	if (test_mode) {
	/*
		PbmsConOpts opts;
		
		opts.engine = TEST_ENGINE_NAME;
		opts.host = host;
		opts.user = user;
		opts.passwd = password;
		opts.port = port;
		
		stress_test_api(mysql, timeout, &opts);
	*/
		printf("Not implements.\n");
	} else 
		test_api(mysql);

	// Clean up
	done:	
	
	if (results)
		mysql_free_result(results);

	if (mysql)
		mysql_close(mysql);
		
	mysql_library_end();
	
    return 0;
}
