/* Copyright (c) 2009 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase Media Stream for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Barry Leslie
 *
 * 2009-07-31
 *
 * H&G2JCtL
 *
 * PHP module for the PrimeBase Media Stream (PBMS) daemon.
 *
 */
#include "php.h"
#include "php_ini.h"

#include "ext/standard/info.h"
#include "pbmslib.h"
#include "php_pbms.h"
#include "php_streams.h"

#ifdef ZTS
# define PBMS_LS_D	zend_pbms_globals *pbms_globals
# define PBMS_LS_DC	, PBMS_LS_D
# define PBMS_LS_C	pbms_globals
# define PBMS_LS_CC , PBMS_LS_C
# define PBMS_G(v) (pbms_globals->v)
# define PBMS_LS_FETCH()	zend_pbms_globals *pbms_globals = ts_resource(pbms_globals_id)
#else
# define PBMS_LS_D
# define PBMS_LS_DC
# define PBMS_LS_C
# define PBMS_LS_CC
# define PBMS_G(v) (pbms_globals.v)
# define PBMS_LS_FETCH()
#endif

#ifndef TRANSFER_BUFFER_SIZE
#define TRANSFER_BUFFER_SIZE (10 * 1024)
#endif

#define PBMS_CONNECTION_NAME	"PrimeBase pbms connection"
#define PBMS_POOL_NAME			"PrimeBase pbms pool"
#define PBMS_MAX_HOST_SIZE		124	// Maximum length of an IP address allowed.
#define PBMS_NAME_SIZE			64		
#define MAX_ERROR				256

#define	PBMS_BASE_ERROR			5000
#define PHP_PBMS_CONID_ERR	(PBMS_BASE_ERROR + 1)
#define PHP_PBMS_CALLBACK_ERR (PBMS_BASE_ERROR + 2)
#define PHP_PBMS_MEMORY_ERR (PBMS_BASE_ERROR + 3)

#define SAFE_STRING(s) ((s)?(s):"")


typedef unsigned char Bool;
#undef TRUE
#undef FALSE
#undef OK
#undef FAILED
#define TRUE	1
#define FALSE	0
#define OK		TRUE
#define FAILED	FALSE

typedef struct {
	PBMS	pbms;	

	char	host[PBMS_MAX_HOST_SIZE];
	long	port;
	char	database[PBMS_NAME_SIZE];	// A list of open databases.	
	
	Bool	display_error; 		// Report the error immediatly.
	char	err_msg[MAX_ERROR+1];	// Error text.
	int		err_no;
	
	Bool	in_use; 
	struct PBMSPool_tag *my_pool; 
} PBMSCon, *PBMSConPtr;

#define CLEAR_ERRORS(pbcon) {pbcon->err_no = 0;pbcon->err_msg[0] = 0;}

#ifndef MAX_PBMS_POOL_CONNECTIONS
#define MAX_PBMS_POOL_CONNECTIONS 100
#endif

typedef struct PBMSPool_tag{
	int pool_size;	// The number of connections in the list. 
	PBMSConPtr	con_list[MAX_PBMS_POOL_CONNECTIONS];
	struct PBMSPool_tag *next;
	struct PBMSPool_tag *prev;
} PBMSPool, *PBMSPoolPtr;

static PBMSPoolPtr pools = NULL;
static int le_connection, le_pool;

ZEND_DECLARE_MODULE_GLOBALS(pbms)

//------------------------------------------------------
PHP_INI_BEGIN()
	STD_PHP_INI_BOOLEAN("pbms.display_errors",	"1",		PHP_INI_SYSTEM,		OnUpdateLong,		display_errors,		zend_pbms_globals,		pbms_globals)
	STD_PHP_INI_BOOLEAN("pbms.debug_mode",		"0",		PHP_INI_SYSTEM,		OnUpdateLong,		debug_mode,			zend_pbms_globals,		pbms_globals)
	STD_PHP_INI_ENTRY_EX("pbms.max_connections","-1",		PHP_INI_SYSTEM,		OnUpdateLong,		max_connections,	zend_pbms_globals,		pbms_globals,	display_link_numbers)
	STD_PHP_INI_ENTRY_EX("pbms.num_connections","0",		PHP_INI_SYSTEM,		OnUpdateLong,		num_connections,	zend_pbms_globals,		pbms_globals,	display_link_numbers)
	STD_PHP_INI_ENTRY_EX("pbms.pool_size",		"0",		PHP_INI_SYSTEM,		OnUpdateLong,		pool_size,			zend_pbms_globals,		pbms_globals,	display_link_numbers)

	STD_PHP_INI_ENTRY("pbms.default_host",		"localhost",PHP_INI_ALL,		OnUpdateString,		default_host,		zend_pbms_globals,		pbms_globals)
	STD_PHP_INI_ENTRY("pbms.default_port",		"8080",		PHP_INI_ALL,		OnUpdateLong,		default_port,		zend_pbms_globals,		pbms_globals)
	STD_PHP_INI_ENTRY("pbms.default_database",	"",			PHP_INI_ALL,		OnUpdateString,		default_database,	zend_pbms_globals,		pbms_globals)
PHP_INI_END()

//------------------------------------------------------
static void php_pbms_init_globals(zend_pbms_globals *pbms_globals)
{
#ifdef ZTS
	pool_mutex = tsrm_mutex_alloc();
#endif
}

//------------------------------------------------------
static void lock_pools(void)
{
#ifdef ZTS
	tsrm_mutex_lock(pool_mutex);
#endif
}

//------------------------------------------------------
static void unlock_pools(void)
{
#ifdef ZTS
	tsrm_mutex_unlock(pool_mutex);
#endif
}


//------------------------------------------------------
static void _free_pbms_pool(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	PBMSPoolPtr pool = (PBMSPoolPtr) rsrc->ptr;
	PBMSConPtr pbcon;
	int i;
	PBMS_LS_FETCH();

	lock_pools();
	for (i = 0; i < pool->pool_size; i++) {
		pbcon = pool->con_list[i];
		if (pbcon) {
			if ( pbcon->pbms && !pbcon->in_use) {
				PBMS_G(num_connections)--;
				pbms_close(pbcon->pbms);
				free(pbcon);
			} else 
				pbcon->my_pool = NULL;
		}
	}
	
	if (pools == pool) 	
		pools = pool->next;
	else
		pool->prev->next = pool->next;
		
	if (pool->next)
		pool->next->prev = pool->prev;
		
	free(pool);
	unlock_pools();
}

//------------------------------------------------------
static void _close_pbms_conn(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	PBMSConPtr pbcon = (PBMSConPtr) rsrc->ptr;
	PBMSPoolPtr pool;
	int i;
	PBMS_LS_FETCH();

	lock_pools();
	pool = pbcon->my_pool;
	if (!pool) { // If the connection isn't in a pool free it.
		PBMS_G(num_connections)--;
		pbms_close(pbcon->pbms);
		free(pbcon);
	} else {
		pbcon->in_use = FALSE;
		if (pool->pool_size > PBMS_G(pool_size)) {
			PBMS_G(num_connections)--;
			pbms_close(pbcon->pbms);
			pbcon->pbms = NULL;
			free(pbcon);
			for (i = 0; (i < pool->pool_size) && (pool->con_list[i] != pbcon); i++);
			if (i < pool->pool_size) {
				memmove(pool->con_list +i, pool->con_list +i+1, sizeof(PBMSConPtr) * ( pool->pool_size -i -1));
				pool->pool_size--;
				pool->con_list[pool->pool_size] = NULL;
			}
		}
	}
	

	unlock_pools();
	
}


//------------------------------------------------------
ZEND_MINIT_FUNCTION(pbms)
{
	ZEND_INIT_MODULE_GLOBALS(pbms, php_pbms_init_globals, NULL);
	REGISTER_INI_ENTRIES();

	le_connection = zend_register_list_destructors_ex(_close_pbms_conn, NULL, PBMS_CONNECTION_NAME, module_number);
	le_pool = zend_register_list_destructors_ex(NULL, _free_pbms_pool, PBMS_POOL_NAME, module_number);

    return (pbms_library_init())? SUCCESS: FAILURE;
}

ZEND_MSHUTDOWN_FUNCTION(pbms)
{
	pbms_library_end();
	UNREGISTER_INI_ENTRIES();
    return SUCCESS;
}

ZEND_RINIT_FUNCTION(pbms)
{
	PBMS_LS_FETCH();
	
	PBMS_G(default_connection_id)=-1;
	PBMS_G(num_connections) = 0;
    return SUCCESS;
}

ZEND_RSHUTDOWN_FUNCTION(pbms)
{
    return SUCCESS;
}

PHP_MINFO_FUNCTION(pbms)
{
	char buf[32];
	PBMS_LS_FETCH();

	php_info_print_table_start();
	php_info_print_table_header(2, "PrimeBase Media Streaming (PBMS) Support", "enabled");
	snprintf(buf, 32, "%d", PBMS_G(pool_size));
	php_info_print_table_row(2, "Persistent connection pool size", buf);
	snprintf(buf, 32, "%ld", PBMS_G(num_connections));
	php_info_print_table_row(2, "Active Connections", buf);

    php_info_print_table_end();
}

//------------------------------------------------------------------------------------
static void report_pbms_error(PBMSConPtr pbcon)
{	
	
	pbcon->err_no = pbms_errno(pbcon->pbms);
	strncpy(pbcon->err_msg, pbms_error(pbcon->pbms), MAX_ERROR);
	pbcon->err_msg[MAX_ERROR] = 0;
	
	if (pbcon->display_error)
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", pbcon->err_msg);	
}

//------------------------------------------------------------------------------------
static void report_php_error(PBMSConPtr pbcon, int err, const char *msg)
{
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", msg);	
		return;
	}
	
	pbcon->err_no = err;
	strncpy(pbcon->err_msg, msg, MAX_ERROR);
	pbcon->err_msg[MAX_ERROR] = 0;
	
	if (pbcon->display_error)
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", pbcon->err_msg);	
}


//------------------------------------------------------------------------------------
static Bool set_connection(PBMSConPtr pbcon)
{

	if (pbcon->pbms)
		pbms_close(pbcon->pbms);
		
	pbcon->pbms = pbms_connect(pbcon->host, pbcon->port, pbcon->database);
	if (!pbcon->pbms) {
		report_pbms_error(pbcon);
		return FAILED;
	}
	
	return OK;
	
}


//------------------------------------------------------------------------------------
// I do not understand how INTERNAL_FUNCTION_PARAMETERS works and so I do not trust it and therefore do not use it.
// If I ever found some actual documentation saying how it should and should not be used then I may trust it.
// The reason I use it is so that I can use the return value setting macros.
// I hate macros and Zend is macro madness. :(
static void php_pbms_do_connect(INTERNAL_FUNCTION_PARAMETERS, char *host, long port, char *database, int pool_connection)
{
	char *hashed_details;
	int i, hashed_details_length;
  	PBMSConPtr pbcon;
	list_entry *pool_list;
	PBMSPoolPtr pool = NULL;
	
	PBMS_LS_FETCH();

	if ((!host) || !*host) 
		host = PBMS_G(default_host);
	
	if (!port) 
		port = (long)PBMS_G(default_port);
	
	if (!database)
		database = PBMS_G(default_database);
		
	hashed_details_length = strlen("pbmsengine___") + strlen(SAFE_STRING(host)) + strlen(SAFE_STRING(database))+(sizeof(port)*2);
	hashed_details = (char *) emalloc(hashed_details_length+1);
	if (!hashed_details) {
		php_error(E_WARNING, "Out Of Memory!");
		RETURN_FALSE;
	}
	snprintf(hashed_details, hashed_details_length, "pbmsengine_%s_%s_%x", SAFE_STRING(host), SAFE_STRING(database), port);
	
	/* try to find if we already have this connection in our persistent list */
	if ((!PBMS_G(pool_size)) || zend_hash_find(&EG(persistent_list), hashed_details, hashed_details_length+1, (void **) &pool_list)==FAILURE) {  /* we don't */

		if (PBMS_G(max_connections)!=-1 && PBMS_G(num_connections)>=PBMS_G(max_connections)) {
			php_error(E_WARNING, "PBMS:  Too many open connections (%ld)", PBMS_G(num_connections));
			efree(hashed_details);
			RETURN_FALSE;
		}
		
		/* Open the connection */
		pbcon = (PBMSConPtr) malloc(sizeof(PBMSCon));
		if (!pbcon) {
			php_error(E_WARNING, "Out Of Memory!");
			efree(hashed_details);
			RETURN_FALSE;
		}
		memset(pbcon, 0, sizeof(PBMSCon));
		pbcon->display_error =  PBMS_G(display_errors);
		
		if (pool_connection && PBMS_G(pool_size)) {
			pool = (PBMSPoolPtr) malloc(sizeof(PBMSPool));
			if (!pool) {
				php_error(E_WARNING, "Out Of Memory!");
				efree(hashed_details);
				free(pbcon);
				RETURN_FALSE;
			}
			memset(pool, 0, sizeof(PBMSPool));
		}
		
		if (host)
			strncpy(pbcon->host, host, PBMS_MAX_HOST_SIZE);
		if (database) 
			strcpy(pbcon->database, database);
	
		pbcon->port = port;

		if (!set_connection(pbcon)) {
			free(pbcon);
			if (pool) free(pool);
			efree(hashed_details);
			RETURN_FALSE;
		}
		PBMS_G(num_connections)++;
		
		
		pbcon->in_use = TRUE;
		pbcon->my_pool = pool;

		if (pool) {
			list_entry new_pool;
			
			pool->con_list[0] = pbcon;
			pool->pool_size = 1;
			
			
			/* hash it up */
			new_pool.type = le_pool;
			new_pool.ptr = pool;
			if (zend_hash_update(&EG(persistent_list), hashed_details, hashed_details_length+1, (void *) &new_pool, sizeof(list_entry), NULL)==FAILURE) {
				PBMS_G(num_connections)--;
				pbms_close(pbcon->pbms);
				free(pbcon);
				free(pool);
				efree(hashed_details);
				RETURN_FALSE;
			}
			
			lock_pools();
			pool->next = pools;
			pools = pool;
			unlock_pools();
			
		}
	} else {  /* we do */
		if (pool_list->type != le_pool) {
			RETURN_FALSE;
		}
		
		pool = (PBMSPoolPtr) pool_list->ptr;
		pbcon = NULL;
		lock_pools();
		for (i = 0; i < pool->pool_size && pool->con_list[i]->in_use; i++);
		if ( i < pool->pool_size) {
			pbcon = pool->con_list[i];
			pbcon->in_use = TRUE;
		}
		unlock_pools();
		
		if (!pbcon) {
			if (PBMS_G(max_connections)!=-1 && PBMS_G(num_connections)>=PBMS_G(max_connections)) {
				php_error(E_WARNING, "PBMS:  Too many open connections (%ld)", PBMS_G(num_connections));
				efree(hashed_details);
				RETURN_FALSE;
			}
			
			if (pool->pool_size == MAX_PBMS_POOL_CONNECTIONS) {
				php_error(E_WARNING, "Maximum connections per server exceeded! (As defined by the compiler flag MAX_PBMS_POOL_CONNECTIONS.)");
				efree(hashed_details);
				RETURN_FALSE;
			}
			
			pbcon = (PBMSConPtr) malloc(sizeof(PBMSCon));
			if (!pbcon) {
				php_error(E_WARNING, "Out Of Memory!");
				efree(hashed_details);
				RETURN_FALSE;
			}
			memset(pbcon, 0, sizeof(PBMSCon));
			pbcon->display_error =  PBMS_G(display_errors);
			
			pbcon->in_use = TRUE;
			if (host)
				strncpy(pbcon->host, host, PBMS_MAX_HOST_SIZE);
				
			if (database) 
				strcpy(pbcon->database, database);
		
			pbcon->port = port;
			
			lock_pools();
			if (pool_connection && (pool->pool_size < PBMS_G(pool_size))) {			
				pbcon->my_pool = pool;
				pool->con_list[pool->pool_size] = pbcon;
				pool->pool_size++;
			} else {
				pbcon->my_pool = NULL;
			}
			unlock_pools();
		}
		
		/* ensure that the link did not die */
		if (!pbcon->pbms){ // The connection died.

			if (!set_connection(pbcon)) {
				pbcon->in_use = FALSE;
				if (!pbcon->my_pool)
					free(pbcon);
				efree(hashed_details);
				RETURN_FALSE;
			}
			
			PBMS_G(num_connections)++;
		}
			
	}
	
	ZEND_REGISTER_RESOURCE(return_value, pbcon, le_connection);

	efree(hashed_details);
	
	if (PBMS_G(default_connection_id)!= -1) {
		zend_list_delete(PBMS_G(default_connection_id)); // Removes a reference to the link
	}
	PBMS_G(default_connection_id) = return_value->value.lval;
	zend_list_addref(return_value->value.lval); // Add a reference so the link isn't deleted prematurely.
}

//------------------------------------------------------------------------------------
static PBMSConPtr php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAMETERS, Bool create)
{
	zval **zpbcon=NULL;
	PBMSConPtr pbcon = NULL;
	int pbms_res_type;
	PBMS_LS_FETCH();

	
	if (PBMS_G(default_connection_id)== -1) { /* no link opened yet, implicitly open one */
		if (!create)
			return NULL;
		php_pbms_do_connect(INTERNAL_FUNCTION_PARAM_PASSTHRU, NULL, 0, NULL, 0);
	}
	
	ZEND_FETCH_RESOURCE(pbcon, PBMSConPtr, zpbcon, PBMS_G(default_connection_id), PBMS_CONNECTION_NAME, le_connection);

	return pbcon;
}

//------------------------------------------------------------------------------------
// long pbms_errno([int connection_identifier])
ZEND_FUNCTION(pbms_errno)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	int err = PHP_PBMS_CONID_ERR;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, FALSE);
	
	if (pbcon) {
		err = pbcon->err_no;
	}
	
   RETURN_LONG(err);
}

//------------------------------------------------------------------------------------
// string pbms_error([int connection_identifier])
ZEND_FUNCTION(pbms_error)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *err_msg = "Invalid PBMS Connection ID";

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, FALSE);
	
	if (pbcon) {
		err_msg = pbcon->err_msg;
	}
	

    RETURN_STRING(err_msg, 1);
}

//------------------------------------------------------------------------------------
// A persistent connection.
// int pbms_pconnect([string host[, long port[, string database]]]])
ZEND_FUNCTION(pbms_pconnect)
{
	char *host = NULL, *database = NULL;
	int host_len, database_len;
	long port = 0;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|sls", &host, &host_len, &port, &database, &database_len)){
        RETURN_FALSE;
	}

	php_pbms_do_connect(INTERNAL_FUNCTION_PARAM_PASSTHRU PBMS_LS_DC, host, port, database, 1);
}

//------------------------------------------------------------------------------------
// A non persistent connection.
// int pbms_connect([string host[, long port[, string database]]]])
ZEND_FUNCTION(pbms_connect)
{
	char *host = NULL, *database = NULL;
	int host_len, database_len;
	long port = 0;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS()  TSRMLS_CC, "|sls", &host, &host_len, &port, &database, &database_len)) {
        RETURN_FALSE;
	}
	
	php_pbms_do_connect(INTERNAL_FUNCTION_PARAM_PASSTHRU PBMS_LS_DC, host, port, database, 0);
}

//------------------------------------------------------------------------------------
// void pbms_close([int connection_identifier])
ZEND_FUNCTION(pbms_close)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	PBMS_LS_FETCH();

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
			RETURN_FALSE;
		}
		
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, FALSE);

	if (pbcon) {
		// In case this is a pool connection clear the metadata so it is ready
		// for the next session.
		pbms_clear_metadata(pbcon->pbms, NULL);

		CLEAR_ERRORS(pbcon);
		if (zpbcon) {
			zend_list_delete(Z_RESVAL_P(zpbcon));
			if (Z_RESVAL_P(zpbcon) == PBMS_G(default_connection_id))
				PBMS_G(default_connection_id) = -1;
		} else {
			zend_list_delete(PBMS_G(default_connection_id));
			PBMS_G(default_connection_id) = -1;
		}
	}
	RETURN_TRUE
}

//------------------------------------------------------------------------------------
// void pbms_add_metadata(string name, string value, [int connection_identifier])
ZEND_FUNCTION(pbms_add_metadata)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
    int name_len;
    int value_len;
    const char *name = NULL;
    const char *value = NULL;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|r", &name, &name_len, &value, &value_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}
	
	CLEAR_ERRORS(pbcon);
	
	if (!pbms_add_metadata(pbcon->pbms, name, value)) {
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}
	

	RETURN_TRUE;
}

//------------------------------------------------------------------------------------
// void pbms_clear_metadata([string name [,int connection_identifier]])
ZEND_FUNCTION(pbms_clear_metadata)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
	int name_len;
	const char *name = NULL;
	int pbms_res_type;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|sr", &name, &name_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}
	
	CLEAR_ERRORS(pbcon);
		
	pbms_clear_metadata(pbcon->pbms, name);
	
    RETURN_TRUE;
}


//------------------------------------------------------------------------------------
// long pbms_reset_metadata([int connection_identifier])
ZEND_FUNCTION(pbms_reset_metadata)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
       RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}
	
	CLEAR_ERRORS(pbcon);

    RETURN_LONG(pbms_reset_metadata(pbcon->pbms));
}

//------------------------------------------------------------------------------------
// (Name,Value) pbms_next_metadata([int connection_identifier])
ZEND_FUNCTION(pbms_next_metadata)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char name[MS_META_NAME_SIZE];
	char *value = NULL;
	size_t v_size, size;
	Bool has_metadata;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

#ifdef DEBUG
	v_size = 5;
#else
	v_size = 128;
#endif

	do {
		if (value)
			value = (char*) erealloc(value, v_size);
		else
			value = (char*) emalloc(v_size);
		if (!value) {
			report_php_error(pbcon, PHP_PBMS_MEMORY_ERR, "Out of memory");
			RETURN_FALSE;
		}
		
		size = v_size;
		has_metadata = pbms_next_metadata(pbcon->pbms, name, value, &v_size);
	} while (size < v_size);

	if (!has_metadata) {
		if (pbms_errno(pbcon->pbms)) {
			report_pbms_error(pbcon);
			RETURN_FALSE;
		}
		*name = 0;
		*value = 0;
	}
	
	// Setup the return value as an array.
	array_init(return_value);
	add_next_index_string(return_value, name, 1);
	add_next_index_string(return_value, value, 0);
	
}

//------------------------------------------------------------------------------------
// string pbms_get_metadata_value(string name[, int connection_identifier])
ZEND_FUNCTION(pbms_get_metadata_value)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *name;
    int name_len;
	char *value = NULL;
	size_t v_size, size;
	Bool has_metadata;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &name, &name_len, &zpbcon)){
       RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

#ifdef DEBUG
	v_size = 5;
#else
	v_size = 128;
#endif

	do {
		if (value)
			value = (char*) erealloc(value, v_size);
		else
			value = (char*) emalloc(v_size);
		if (!value) {
			report_php_error(pbcon, PHP_PBMS_MEMORY_ERR, "Out of memory");
			RETURN_FALSE;
		}
		
		size = v_size;
		has_metadata = pbms_get_metadata_value(pbcon->pbms, name, value, &v_size);
	} while (size < v_size);

	if (!has_metadata) {
		if (pbms_errno(pbcon->pbms)) {
			report_pbms_error(pbcon);
			RETURN_FALSE;
		}
		*value = 0;
	}

    RETURN_STRING(value, 0);
}

//------------------------------------------------------------------------------------
// string pbms_get_md5_digest([, int connection_identifier])
ZEND_FUNCTION(pbms_get_md5_digest)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *md5_digest;
 
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	md5_digest = (char *) emalloc(16);
	if (!pbms_get_md5_digest(pbcon->pbms, md5_digest)) {
		efree(md5_digest);
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}
	
	return_value->value.str.len = 16;
   	return_value->value.str.val = md5_digest;
   	return_value->type = IS_STRING;	
}


//------------------------------------------------------------------------------------
// void pbms_get_info(string blob_ref [, int connection_identifier])
ZEND_FUNCTION(pbms_get_info)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
  
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (!pbms_get_info(pbcon->pbms, blob_ref)) {
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}
	
    RETURN_TRUE;
}

typedef struct {
	php_stream *stream;
	char *error_msg;
} CBDataRec, *CBDataPtr;

//---------------------------------	
static size_t my_stream_write(void *caller_data, const char *buffer, size_t size, Bool reset)
{
	CBDataPtr data = (CBDataPtr) caller_data;
	size_t cnt;
	
	if (reset) {
		if (php_stream_seek(data->stream, 0, SEEK_SET)) {
			data->error_msg = "PBMS: my_stream_write(): php_stream_seek() error";
			return 0;
		}
	}
	cnt = php_stream_write(data->stream, buffer, size);
	if (cnt < size) 
		data->error_msg = "PBMS: my_stream_write(): php_stream_write() error";
		
	return cnt;
}


// void pbms_write_stream(php_stream *output_stream, string blob_ref [, int connection_identifier])
ZEND_FUNCTION(pbms_write_stream)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
	char buffer[TRANSFER_BUFFER_SIZE];
	zval *zstream;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
 	CBDataRec cbData = {0};
 
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zs|r", &zstream, &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}


	CLEAR_ERRORS(pbcon);

    php_stream_from_zval(cbData.stream, &zstream);
	cbData.error_msg = NULL;
	
	if ((!pbms_get_data_cb(pbcon->pbms, blob_ref, my_stream_write, &cbData)) || cbData.error_msg) {
		if (cbData.error_msg) 
			report_php_error(pbcon, PHP_PBMS_CALLBACK_ERR, cbData.error_msg);
		else
			report_pbms_error(pbcon);
		RETURN_FALSE;
	}

    RETURN_TRUE;
}

// void pbms_write_stream_range(php_stream *output_stream, string blob_ref, long first_byte, long last_byte [, int connection_identifier])
ZEND_FUNCTION(pbms_write_stream_range)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
	char buffer[TRANSFER_BUFFER_SIZE];
	zval *zstream;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
	long first_byte, last_byte;
 	CBDataRec cbData = {0};
 
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zsll|r", &zstream, &blob_ref, &blob_ref_len, &first_byte, &last_byte, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	if (first_byte > last_byte) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Range Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

    php_stream_from_zval(cbData.stream, &zstream);
	cbData.error_msg = NULL;
	
	if ((!pbms_get_data_range_cb(pbcon->pbms, blob_ref, first_byte, last_byte, my_stream_write, &cbData)) || cbData.error_msg) {
		if (cbData.error_msg) 
			report_php_error(pbcon, PHP_PBMS_CALLBACK_ERR, cbData.error_msg);
		else
			report_pbms_error(pbcon);
		RETURN_FALSE;
	}

    RETURN_TRUE;
}

// string pbms_get_data(string blob_ref [, int connection_identifier])
ZEND_FUNCTION(pbms_get_data)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	unsigned char * buffer;
	int blob_ref_len;
	size_t size;
  
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (!pbms_get_blob_size(pbcon->pbms, blob_ref, &size)){
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}

	buffer = (unsigned char *) emalloc(size);
	if (!buffer) {
		report_php_error(pbcon, PHP_PBMS_MEMORY_ERR, "Out of memory");
		RETURN_FALSE;
	}
	
	if (!pbms_get_data(pbcon->pbms, blob_ref, buffer, size)) {
		report_pbms_error(pbcon);
		efree(buffer);
		RETURN_FALSE;
	}
	
	return_value->value.str.len = size;
   	return_value->value.str.val = (char*) buffer;
   	return_value->type = IS_STRING;
}

//---------------------------------	
// string pbms_get_data_range(string blob_ref, long first_byte, long last_byte [, int connection_identifier])
ZEND_FUNCTION(pbms_get_data_range)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	unsigned char * buffer;
	int blob_ref_len;
	size_t size, data_size =0;
  	long first_byte, last_byte;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sll|r", &blob_ref, &blob_ref_len, &first_byte, &last_byte, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	if (first_byte > last_byte) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Range Provided");
		RETURN_FALSE;
	}
	
	CLEAR_ERRORS(pbcon);

	if (!pbms_get_blob_size(pbcon->pbms, blob_ref, &size)){
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}

	if (size > (last_byte - first_byte))
		size = last_byte - first_byte +1;
		
	buffer = (unsigned char *) emalloc(size);
	if (!buffer) {
		report_php_error(pbcon, PHP_PBMS_MEMORY_ERR, "Out of memory");
		RETURN_FALSE;
	}
	
	if (!pbms_get_data_range(pbcon->pbms, blob_ref, first_byte, last_byte, buffer, size, &data_size)) {
		report_pbms_error(pbcon);
		efree(buffer);
		RETURN_FALSE;
	}
	
	return_value->value.str.len = data_size;
   	return_value->value.str.val = (char*) buffer;
   	return_value->type = IS_STRING;
}

//---------------------------------	
static size_t my_stream_read(void *caller_data, char *buffer, size_t size, Bool reset)
{
	CBDataPtr data = (CBDataPtr) caller_data;
	size_t cnt;
	
	if (reset) {
		if (php_stream_seek(data->stream, 0, SEEK_SET)) {
			data->error_msg = "PBMS: my_stream_read(): php_stream_seek() error";
			return 0;
		}
	}

	cnt = php_stream_read(data->stream, buffer, size);
	if ((cnt < size) && (php_stream_eof(data->stream) != 1)) {
		data->error_msg = "PBMS: my_stream_read(): php_stream_read() error";
	}
		
	return cnt;
}


// string pbms_read_stream(php_stream *input_stream, int size,  [, string table[, string checksum[, int connection_identifier]]])
ZEND_FUNCTION(pbms_read_stream)
{
	PBMSConPtr pbcon;
	CBDataRec cbData = {0};
	php_stream *stream;
    zval *zpbcon = NULL;
	char buffer[TRANSFER_BUFFER_SIZE];
	char blob_ref[MS_BLOB_URL_SIZE];
	zval *zstream;
	zval *ztable = NULL;
	zval *zchecksum = NULL;
	char *table = NULL;
	char *checksum = NULL;
	
	int pbms_res_type;
	int size;
    
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl|z!z!r", &zstream, &size, &ztable, &zchecksum, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

    php_stream_from_zval(cbData.stream, &zstream);
	cbData.error_msg = NULL;
	
	if (ztable)
		table = Z_STRVAL_P(ztable);
		
 	if (zchecksum)
		checksum = Z_STRVAL_P(zchecksum);

	if ((!pbms_put_data_cb(pbcon->pbms, table, checksum, blob_ref, size, my_stream_read, &cbData)) || cbData.error_msg) {
		if (cbData.error_msg) 
			report_php_error(pbcon, PHP_PBMS_CALLBACK_ERR, cbData.error_msg);
		else
			report_pbms_error(pbcon);
		RETURN_FALSE;
	}

    RETURN_STRING(blob_ref, 1);
}

// string pbms_put_data( string data,  [, string table[, string checksum [, int connection_identifier]]])
ZEND_FUNCTION(pbms_put_data)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
	char blob_ref[MS_BLOB_URL_SIZE];
	zval *ztable = NULL;
	zval *zchecksum = NULL;
	char *table = NULL;
	char *checksum = NULL;
	unsigned char *data = NULL;	
	int pbms_res_type;
	int size;
       
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|z!z!r", &data, &size, &ztable, &zchecksum, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (ztable)
		table = Z_STRVAL_P(ztable);
		
 	if (zchecksum)
		checksum = Z_STRVAL_P(zchecksum);

	if (!pbms_put_data(pbcon->pbms, table, checksum, blob_ref, size, data)) {
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}

    RETURN_STRING(blob_ref, 1);
}

// bool pbms_is_blob_reference( string blob_ref[, int connection_identifier])
ZEND_FUNCTION(pbms_is_blob_reference)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
	Bool is_ref = FALSE;
  
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (!pbms_is_blob_reference(pbcon->pbms, blob_ref)) {
		if (pbms_errno(pbcon->pbms)) {
			report_pbms_error(pbcon);
			RETURN_FALSE;
		}
	} else
		is_ref = TRUE;

    RETVAL_BOOL(is_ref);
}

// long pbms_get_blob_size( string blob_ref[, int connection_identifier])
ZEND_FUNCTION(pbms_get_blob_size)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
	size_t size;
  
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (!pbms_get_blob_size(pbcon->pbms, blob_ref, &size)) {
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}
	
    RETURN_LONG(size);
}

#if ZEND_MODULE_API_NO >= 20050922
static zend_module_dep pbms_module_deps[] = {
    ZEND_MOD_REQUIRED("mysql")
    {NULL, NULL, NULL}
};
#endif

zend_module_entry pbms_module_entry = {
#if ZEND_MODULE_API_NO >= 20050922
	STANDARD_MODULE_HEADER_EX, NULL,
    pbms_module_deps,
#else
	STANDARD_MODULE_HEADER,
#endif
	PHP_PBMS_EXTNAME,	    	// PHP extension name (PBMS)
	php_pbms_functions,		    // Functions
	ZEND_MINIT(pbms),	        // MINIT
	ZEND_MSHUTDOWN(pbms),		// MSHUTDOWN
	ZEND_RINIT(pbms),		    // RINIT
	ZEND_RSHUTDOWN(pbms),	    // RSHUTDOWN
	PHP_MINFO(pbms),			// MINFO
#if ZEND_MODULE_API_NO >= 20010901
	PHP_PBMS_EXTVER,
#endif
	STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_PBMS
	ZEND_GET_MODULE(pbms)		// for PHP extensions built as shared modules
#endif
