--TEST--
Basic transaction tests involving PBMS BLOBs.
--SKIPIF--
<?php if (!extension_loaded("PBMS")) print "skip: PBMS not loaded."; ?>
--POST--
--GET--
--FILE--
<?php 
	include "pbms_utils.php";

	/*
	 * Note on testing PBMS transactions:
	 * Any BLOB that is inserted, deleted, or updated will appear in the pbms_reference table
	 * as a referenced BLOB until the transaction is committed or rolled back. Commiting a
	 * delete or rolling back an insert will dereference the BLOB and set it's 'Deletion_time'
	 * to a non NULL value.
	 *
	 * Because the BLOB is never scheduled for deletion until the transaction has completed
	 * means that the PBMS transaction manager can do things at it's own pace so in theory the
	 * the timing of when the BLOBs are scheduled for deletion is not guarenteed. This is not a problem
	 * except for in tests such as this one. You should be aware of this when writing or debugging
	 * PBMS transaction tests.
	 * 
	 */
	 
	
	print "Basic transaction tests involving PBMS BLOBs:\n";
	pbms_connect($host, $port, $database)
		or pbms_error_and_die("Could not connect\n");
		
	// Be sure to actually opendifferent sessions!
	$mysql1 = mysql_connect( ini_get("mysql.default_host"), ini_get("mysql.default_user"), ini_get("mysql.default_password"), TRUE)
		or mysql_error_and_die("Could not connect\n");

	$mysql2 = mysql_connect( ini_get("mysql.default_host"), ini_get("mysql.default_user"), ini_get("mysql.default_password"), TRUE)
		or mysql_error_and_die("Could not connect\n");

		
	print "Connected successfully\n";
	init_php_test_database($mysql1, $engine, $database);
	
	// Be sure to actually open a different session!
	$mysql2 = mysql_connect( ini_get("mysql.default_host"), ini_get("mysql.default_user"), ini_get("mysql.default_password"), TRUE)
		or mysql_error_and_die("Could not connect\n");

	$query = "Use $database;";
	mysql_query($query, $mysql1)
	or mysql_error_and_die("$query  failed\n");
	
	mysql_query($query, $mysql2)
	or mysql_error_and_die("$query failed\n");
	
	if (mysql_query("Delete from test1; ", $mysql1) == FALSE) {
		mysql_error_and_die("Delete from test1 Failed: \n");
	}

	if (mysql_query("Delete from test2; ", $mysql1) == FALSE) {
		mysql_error_and_die("Delete test2 test1 Failed: \n");
	}

	if (mysql_query("drop table if exists non_existent_table ; ", $mysql1) == FALSE) {
		mysql_error_and_die("drop table if exists non_existent_table : \n");
	}
	// Test auto commit:
	//----------------
	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE);

	$query = "rollback;"; // Should have no effect.
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", TRUE);
	
	$query = "Delete from test1;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE);
	
	print "Test auto commit successfull\n";
	
	// Test basic transaction commit:	
	//----------------
	
	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // There should not be scheduled for deletion even though it was not committed.
	
	$query = "commit;"; 
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); 

	$query = "rollback;"; // Should have no effect.
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); 

	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Delete from test1;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // The BLOB isn't scheduled for deletion until the delete is committed.
	
	$query = "commit;"; 
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE); // The BLOB should now be scheduled for deletion.

	$query = "rollback;"; // Should have no effect.
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE); 

	print "Test basic transaction commit successfull\n";
	
	// Test basic transaction rollback:
	//----------------
	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // There should not be scheduled for deletion even though it was not committed.
	
	$query = "rollback;"; 
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE); // The BLOB should now be scheduled for deletion.

	$query = "commit;"; // Should have no effect.
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE); 

	print "Test basic transaction rollback successfull\n";
	
	// Test statement rollback as the result of a duplicate key:
	//----------------
	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // There should not be scheduled for deletion even though it was not committed.
	
	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	; // Expect duplicate key error.
	mysql_query($query, $mysql1);

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // The BLOB should NOT be scheduled for deletion.

	$query = "commit;"; 
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); 
	if (getReferencedBlobCount($mysql1, "test1") != 1) 
		die("Unexpected number of referenced BLOBs in table.");

	$query = "Delete from test1;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE); 

	// Do the same thing but with  a rollback at the end.
	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // There should not be scheduled for deletion even though it was not committed.
	
	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	; // Expect duplicate key error.
	mysql_query($query, $mysql1);

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // The BLOB should NOT be scheduled for deletion.

	$query = "rollback;"; 
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE); 
	if (getReferencedBlobCount($mysql1, "test1") != 0) 
		die("Unexpected referenced BLOBs in table.");

	
	print "Test statement rollback as the result of a duplicate key successfull\n";	
	
	// Test commit as the result of a DD statement:
	//----------------
	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // There should not be scheduled for deletion even though it was not committed.
	
	$query = "create table non_existent_table(c1 int);"; 
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "drop table non_existent_table;"; 
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // The transaction should have been committed.

	$query = "rollback;"; // Should have no effect.
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); 
	
	$query = "Delete from test1;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE); // The BLOB is scheduled for deletion becuase of auto commit.

	print "Test commit as the result of a DD statement successfull\n";	

	// Test rollback to save point:
	//----------------
	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$query = "savepoint pbmstest1;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$query = "Insert into test1(c1, c2) values( 2, 'An other test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id1 = getBlobID(1, "test1", "c2", $mysql1);
	$blob_id2 = getBlobID(2, "test1", "c2", $mysql1);
	
	testBlobStatus($mysql1, $blob_id1, "test1", TRUE); // Should not be scheduled for deletion.
	testBlobStatus($mysql1, $blob_id2, "test1", TRUE); // Should not be scheduled for deletion.
	
	$query = "Rollback to pbmstest1;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Commit;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
		
	testBlobStatus($mysql1, $blob_id1, "test1", TRUE); // Should not be scheduled for deletion.
	testBlobStatus($mysql1, $blob_id2, "test1", FALSE); // Should be scheduled for deletion.
	
	$query = "Delete from test1 where c1 = 1;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
		
	testBlobStatus($mysql1, $blob_id1, "test1", FALSE); // Should be scheduled for deletion.
	testBlobStatus($mysql1, $blob_id2, "test1", FALSE); // Should be scheduled for deletion.

	print "Test rollback to save point successfull\n";	

	// Test transactions involving updates:
	//----------------
	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE);
	 
	$query = "update test1 set c2 = 'A different test blob' where c1 = 1;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id2 = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", FALSE); // The old blob should be scheduled for deletion.
	testBlobStatus($mysql1, $blob_id2, "test1", TRUE);
	
	$query = "delete from test1;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE);
	 
	$query = "begin;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "update test1 set c2 = 'A different test blob' where c1 = 1;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id2 = getBlobID(1, "test1", "c2", $mysql1);

	$query = "rollback;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", TRUE); // The old blob should NOT be scheduled for deletion.
	testBlobStatus($mysql1, $blob_id2, "test1", FALSE); // The new blob shouldbe scheduled for deletion becaause of rollback.

	$query = "delete from test1;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "begin;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id, "test1", TRUE);
	 
	$query = "update test1 set c2 = 'A different test blob' where c1 = 1;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id2 = getBlobID(1, "test1", "c2", $mysql1);

	$query = "commit;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	testBlobStatus($mysql1, $blob_id, "test1", FALSE); // The old blob should be scheduled for deletion.
	testBlobStatus($mysql1, $blob_id2, "test1", TRUE);

	$query = "delete from test1;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	// Test with duplicate key update error.
	$query = "begin;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id1 = getBlobID(1, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id1, "test1", TRUE);
	 
	$query = "Insert into test1(c1, c2) values( 2, 'A second test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id2 = getBlobID(2, "test1", "c2", $mysql1);
	testBlobStatus($mysql1, $blob_id2, "test1", TRUE);
	 
	$query = "update test1 set c2 = 'A different test blob', c1 = 1 where c1 = 2;"	;
	mysql_query($query, $mysql1); // Expect a duplicate key error;
	
	$query = "commit;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	testBlobStatus($mysql1, $blob_id1, "test1", TRUE);
	testBlobStatus($mysql1, $blob_id2, "test1", TRUE);

	if (getReferencedBlobCount($mysql1, "test1") != 2) 
		die("Unexpected referenced BLOBs in table, expected 2.");

	$query = "delete from test1 where c1 = 1;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	testBlobStatus($mysql1, $blob_id1, "test1", FALSE);
	testBlobStatus($mysql1, $blob_id2, "test1", TRUE);

	if (getReferencedBlobCount($mysql1, "test1") != 1) 
		die("Unexpected referenced BLOBs in table, expected 1.");

	$query = "delete from test1 where c1 = 2;"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	testBlobStatus($mysql1, $blob_id1, "test1", FALSE);
	testBlobStatus($mysql1, $blob_id2, "test1", FALSE);

	if (getReferencedBlobCount($mysql1, "test1") != 0) 
		die("Unexpected referenced BLOBs in table, expected 0.");

	print "Test transactions involving updates successfull\n";	

	// Test transactions on different connections:
	//----------------
	

	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	mysql_query($query, $mysql2)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$query = "Insert into test2(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql2)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id1 = getBlobID(1, "test1", "c2", $mysql1);
	$blob_id2 = getBlobID(1, "test2", "c2", $mysql2);

	testBlobStatus($mysql1, $blob_id1, "test1", TRUE); // Should not be scheduled for deletion.
	testBlobStatus($mysql2, $blob_id2, "test2", TRUE); // Should not be scheduled for deletion.
	
	$query = "Commit;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	testBlobStatus($mysql1, $blob_id1, "test1", TRUE); // Should not be scheduled for deletion.
	testBlobStatus($mysql2, $blob_id2, "test2", TRUE); // Should not be scheduled for deletion.
	
	$query = "Rollback;";
	mysql_query($query, $mysql2)
		or mysql_error_and_die("Query Failed: $query");
	
	testBlobStatus($mysql1, $blob_id1, "test1", TRUE); // Should not be scheduled for deletion.
	testBlobStatus($mysql2, $blob_id2, "test2", FALSE); // Should be scheduled for deletion.
	
	$query = "Delete from test1;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	testBlobStatus($mysql1, $blob_id1, "test1", FALSE); // Should  be scheduled for deletion.

	// Do it again but with rollback and commit in opposite order.
	$query = "Begin;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");

	mysql_query($query, $mysql2)
		or mysql_error_and_die("Query Failed: $query");

	$query = "Insert into test1(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	$query = "Insert into test2(c1, c2) values( 1, 'A test blob');"	;
	mysql_query($query, $mysql2)
		or mysql_error_and_die("Query Failed: $query");
	
	$blob_id1 = getBlobID(1, "test1", "c2", $mysql1);
	$blob_id2 = getBlobID(1, "test2", "c2", $mysql2);

	testBlobStatus($mysql1, $blob_id1, "test1", TRUE); // Should not be scheduled for deletion.
	testBlobStatus($mysql2, $blob_id2, "test2", TRUE); // Should not be scheduled for deletion.
	
	$query = "Rollback;";
	mysql_query($query, $mysql2)
		or mysql_error_and_die("Query Failed: $query");
	
	$query = "Commit;";
	mysql_query($query, $mysql1)
		or mysql_error_and_die("Query Failed: $query");
	
	testBlobStatus($mysql1, $blob_id1, "test1", TRUE); // Should not be scheduled for deletion.
	testBlobStatus($mysql2, $blob_id2, "test2", FALSE); // Should be scheduled for deletion.
	
	$query = "Delete from test1;";
	mysql_query($query, $mysql2)
		or mysql_error_and_die("Query Failed: $query");
	
	testBlobStatus($mysql1, $blob_id1, "test1", FALSE); // Should  be scheduled for deletion.

	print "Test transactions on different connections successfull\n";
?>
--EXPECT--
Basic transaction tests involving PBMS BLOBs:
Connected successfully
Test auto commit successfull
Test basic transaction commit successfull
Test basic transaction rollback successfull
Test statement rollback as the result of a duplicate key successfull
Test commit as the result of a DD statement successfull
Test rollback to save point successfull
Test transactions involving updates successfull
Test transactions on different connections successfull
